import traceback
from ..streams import EndianBinaryReader
from ..helpers import CompressionHelper
import re
from itertools import groupby
from ..enums import ShaderCompilerPlatform, ShaderGpuProgramType, SerializedPropertyType
from ..enums import TextureDimension, PassType

HEADER = '''
//////////////////////////////////////////
//
// NOTE: This is *not* a valid shader file
//
///////////////////////////////////////////
'''[1:]


def export_shader(m_Shader):
    if hasattr(m_Shader, "m_SubProgramBlob"): # 5.3 - 5.4
        decompressedBytes = CompressionHelper.decompress_lz4(
            m_Shader.m_SubProgramBlob, m_Shader.decompressedSize)

        blobReader = EndianBinaryReader(decompressedBytes)
        program = ShaderProgram(blobReader, m_Shader.version)

        return HEADER + program.Export(bytes(m_Shader.m_Script).decode("utf8"))

    if hasattr(m_Shader, "compressedBlob"): # 5.5 and up
        return HEADER + ConvertSerializedShader(m_Shader)

    return HEADER + bytes(m_Shader.m_Script).decode("utf8")

def ConvertSerializedShader(m_Shader):
    shaderPrograms = []

    platformNumber = len(m_Shader.platforms)
    for i in range(platformNumber):
        compressedSize = m_Shader.compressedLengths[i]
        decompressedSize = m_Shader.decompressedLengths[i]

        compressedBytes = m_Shader.compressedBlob[int(m_Shader.offsets[i]):int(m_Shader.offsets[i]) + compressedSize]
        decompressedBytes = CompressionHelper.decompress_lz4(compressedBytes, decompressedSize)

        shaderPrograms.append(ShaderProgram(EndianBinaryReader(decompressedBytes, endian="<"), m_Shader.version))

    return ConvertSerializedShaderParsedForm(m_Shader.m_ParsedForm, m_Shader.platforms, shaderPrograms)

def ConvertSerializedShaderParsedForm(m_ParsedForm, platforms, shaderPrograms):
    sb = []

    sb.append("Shader \"{0}\" {{\n".format(m_ParsedForm.m_Name))

    sb.append(ConvertSerializedProperties(m_ParsedForm.m_PropInfo))

    for m_SubShader in m_ParsedForm.m_SubShaders:
        sb.append(ConvertSerializedSubShader(m_SubShader, platforms, shaderPrograms))

    if m_ParsedForm.m_FallbackName:
        sb.append("Fall back \"{0}\"\n".format(m_ParsedForm.m_FallbackName))

    if m_ParsedForm.m_CustomEditorName:
        sb.append("CustomEditor \"{0}\"\n".format(m_ParsedForm.m_CustomEditorName))

    sb.append("}")
    return "".join(sb)

def ConvertSerializedSubShader(m_SubShader, platforms, shaderPrograms):
    sb = []

    sb.append("SubShader {\n")
    if m_SubShader.m_LOD != 0:
        sb.append(" LOD {0}\n".format(m_SubShader.m_LOD))

    sb.append(ConvertSerializedTagMap(m_SubShader.m_Tags, 1))

    for m_Passe in m_SubShader.m_Passes:
        sb.append(ConvertSerializedPass(m_Passe, platforms, shaderPrograms))

    sb.append("}\n")
    return "".join(sb)

def ConvertSerializedPass(m_Passe, platforms, shaderPrograms):
    sb = []
    if m_Passe.m_Type == PassType.kPassTypeNormal:
        sb.append(" Pass ")
    elif m_Passe.m_Type == PassType.kPassTypeUse:
        sb.append(" UsePass ")
    elif m_Passe.m_Type == PassType.kPassTypeGrab:
        sb.append(" GrabPass ")

    if m_Passe.m_Type == PassType.kPassTypeUse:
        sb.append("\"{0}\"\n".format(m_Passe.m_UseName))
    else:
        sb.append("{\n")

        if m_Passe.m_Type == PassType.kPassTypeGrab:
            if m_Passe.m_TextureName:
                sb.append(" \"{0}\"\n".format(m_Passe.m_TextureName))

        else:
            sb.append(ConvertSerializedShaderState(m_Passe.m_State))

            if len(m_Passe.progVertex.m_SubPrograms) > 0:
                sb.append("Program \"vp\" {\n")
                sb.append(ConvertSerializedSubPrograms(m_Passe.progVertex.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progFragment.m_SubPrograms) > 0:
                sb.append("Program \"fp\" {\n")
                sb.append(ConvertSerializedSubPrograms(m_Passe.progFragment.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progGeometry.m_SubPrograms) > 0:
                sb.append("Program \"gp\" {\n")
                sb.append(ConvertSerializedSubPrograms(m_Passe.progGeometry.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progHull.m_SubPrograms) > 0:
                sb.append("Program \"hp\" {\n")
                sb.append(ConvertSerializedSubPrograms(m_Passe.progHull.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progDomain.m_SubPrograms) > 0:
                sb.append("Program \"dp\" {\n")
                sb.append(ConvertSerializedSubPrograms(m_Passe.progDomain.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

        sb.append("}\n")

    return "".join(sb)

def ConvertSerializedSubPrograms(m_SubPrograms, platforms, shaderPrograms):
    sb = []

    indexFunc = lambda x: x.m_BlobIndex
    typeFunc = lambda x: x.m_GpuProgramType

    groups = groupby(sorted(m_SubPrograms, key=indexFunc), indexFunc)

    for _, _group in groups:
        group = list(_group)
        programs = groupby(sorted(group, key=typeFunc), typeFunc)

        for programKey, _programList in programs:
            subPrograms = list(_programList)
            isTier = len(subPrograms) > 1
            for i in range(len(platforms)):
                platform = platforms[i]

                if CheckGpuProgramUsable(platform, programKey):
                    for subProgram in subPrograms:
                        sb.append("SubProgram \"{0} ".format(GetPlatformString(platform)))

                        if isTier:
                            sb.append("hw_tier{0:02} ".format(subProgram.m_ShaderHardwareTier))

                        sb.append("\" {\n")
                        sb.append(shaderPrograms[i].m_SubPrograms[subProgram.m_BlobIndex].Export())

                        sb.append("\n}\n")

                    break

    return "".join(sb)


def ConvertSerializedShaderState(m_State):
    sb = []
    if m_State.m_Name:
        sb.append(" Name \"{0}\"\n".format(m_State.m_Name))

    if m_State.m_LOD != 0:
        sb.append("  LOD {0}\n".format(m_State.m_LOD))

    sb.append(ConvertSerializedTagMap(m_State.m_Tags, 2))

    sb.append(ConvertSerializedShaderRTBlendState(m_State.rtBlend))

    if m_State.alphaToMask.val > 0:
        sb.append(" AlphaToMask On\n")

    if m_State.zClip and m_State.zClip.val != 1:
        sb.append(" ZClip Off\n")

    if m_State.zTest.val != 4:
        sb.append(" ZTest ")
        if m_State.zTest.val == 0: # kFuncDisabled
            sb.append("Off")
        elif m_State.zTest.val == 1: # kFuncNever
            sb.append("Never")
        elif m_State.zTest.val == 2: # kFuncLess
            sb.append("Less")
        elif m_State.zTest.val == 3: # kFuncEqual
            sb.append("Equal")
        elif m_State.zTest.val == 5: # kFuncGreater
            sb.append("Greater")
        elif m_State.zTest.val == 6: # kFuncNotEqual
            sb.append("NotEqual")
        elif m_State.zTest.val == 7: # kFuncGEqual
            sb.append("GEqual")
        elif m_State.zTest.val == 8: # kFuncAlways
            sb.append("Always")

        sb.append("\n")

    if m_State.zWrite.val != 1: # ZWrite On
        sb.append(" ZWrite Off\n")

    if m_State.culling.val != 2: # Cull Back
        sb.append(" Cull ")
        if m_State.culling.val == 0: # kCullOff
            sb.append("Off")
        elif m_State.culling.val == 1: # kCullFront
            sb.append("Front")

        sb.append("\n")

    if m_State.offsetFactor.val != 0 or m_State.offsetUnits.val != 0:
        sb.append(" Offset {0}, {1}\n".format(m_State.offsetFactor.val, m_State.offsetUnits.val))

    # TODO Stencil

    # TODO Fog

    if m_State.lighting:
        sb.append("  Lighting {0}\n".format(m_State.lighting and "On" or "Off"))

    sb.append("  GpuProgramID {0}\n".format(m_State.gpuProgramID))
    return "".join(sb)


def ConvertSerializedShaderRTBlendState(rbBlend):
    # TODO Blend
    sb = []
    return "".join(sb)

def ConvertSerializedTagMap(m_Tags, intent: int):
    sb = []
    if len(m_Tags.tags) > 0:
        sb.append(" "*intent)
        sb.append("Tags { ")
        for key, value in m_Tags.tags.items():
            sb.append("\"{0}\" = \"{1}\" ".format(key, value))
        sb.append("}\n")

    return "".join(sb)

def ConvertSerializedProperties(m_PropInfo):
    sb = []

    sb.append("Properties {\n")
    for m_Prop in m_PropInfo.m_Props:
        sb.append(ConvertSerializedProperty(m_Prop))

    sb.append("}\n")
    return "".join(sb)

def ConvertSerializedProperty(m_Prop):
    sb = []
    for m_Attribute in m_Prop.m_Attributes:
        sb.append("[{0}] ".format(m_Attribute))

    sb.append("{0} (\"{1}\", ".format(m_Prop.m_Name, m_Prop.m_Description))

    if m_Prop.m_Type == SerializedPropertyType.kColor:
        sb.append("Color")
    elif m_Prop.m_Type == SerializedPropertyType.kVector:
        sb.append("Vector")
    elif m_Prop.m_Type == SerializedPropertyType.kFloat:
        sb.append("Float")
    elif m_Prop.m_Type == SerializedPropertyType.kRange:
        sb.append("Range({0:g}, {1:g})".format(m_Prop.m_DefValue[1], m_Prop.m_DefValue[2]))
    elif m_Prop.m_Type == SerializedPropertyType.kTexture:
        if m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDimAny:
            sb.append("any")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDim2D:
            sb.append("2D")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDim3D:
            sb.append("3D")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDimCUBE:
            sb.append("Cube")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDim2DArray:
            sb.append("2DArray")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDimCubeArray:
            sb.append("CubeArray")

    sb.append(") = ")

    if m_Prop.m_Type in [
        SerializedPropertyType.kColor,
        SerializedPropertyType.kVector
    ]:
        sb.append("({0:g},{1:g},{2:g},{3:g})".format(m_Prop.m_DefValue[0], m_Prop.m_DefValue[1], m_Prop.m_DefValue[2],m_Prop.m_DefValue[3]))
    elif m_Prop.m_Type in [
        SerializedPropertyType.kFloat,
        SerializedPropertyType.kRange
    ]:
        sb.append(m_Prop.m_DefValue[0])
    elif m_Prop.m_Type == SerializedPropertyType.kTexture:
        sb.append("\"{0}\" {{ }}".format(m_Prop.m_DefTexture.m_DefaultName))
    else:
        raise ValueError(m_Prop.m_Type)

    sb.append("\n")
    return "".join(map(str,sb))

def CheckGpuProgramUsable(platform, programType):
    if platform == ShaderCompilerPlatform.kShaderCompPlatformGL:
        return programType == ShaderGpuProgramType.kShaderGpuProgramGLLegacy
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D9:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM20
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM30
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM20
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM30 )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXbox360:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM40
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM40
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM40
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11HullSM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11DomainSM50 )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES20:
        return programType == ShaderGpuProgramType.kShaderGpuProgramGLES
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformNaCl:
        raise NotImplementedError
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformFlash:
        raise NotImplementedError
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11_9x:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramDX10Level9Vertex
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX10Level9Pixel )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES3Plus:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramGLES31AEP
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLES31
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLES3 )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSP2:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPS4:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOne:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSM:
        raise NotImplementedError
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformMetal:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramMetalVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramMetalVS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformOpenGLCore:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramGLCore32
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLCore41
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLCore43 )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformN3DS:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformWiiU:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformVulkan:
        return programType == ShaderGpuProgramType.kShaderGpuProgramSPIRV
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformSwitch:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOneD3D12:
        return ( programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS )
    else:
        raise NotImplementedError


def GetPlatformString(platform):
    if platform == ShaderCompilerPlatform.kShaderCompPlatformGL:
        return "openGL"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D9:
        return "d3d9"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXbox360:
        return "xbox360"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPS3:
        return "ps3"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11:
        return "d3d11"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES20:
        return "gles"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformNaCl:
        return "glesdesktop"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformFlash:
        return "flash"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11_9x:
        return "d3d11_9x"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES3Plus:
        return "gles3"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSP2:
        return "psp2"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPS4:
        return "ps4"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOne:
        return "xboxone"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSM:
        return "psm"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformMetal:
        return "metal"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformOpenGLCore:
        return "glcore"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformN3DS:
        return "n3ds"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformWiiU:
        return "wiiu"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformVulkan:
        return "vulkan"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformSwitch:
        return "switch"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOneD3D12:
        return "xboxone+_d3d12"
    else:
        return "unknown"


class ShaderProgram:
    def __init__(self, reader: EndianBinaryReader, version):
        subProgramCapacity = reader.read_int()
        self.m_SubPrograms = []

        if version >=(2019, 3): # 2019.3 and up
            entrySize = 12
        else:
            entrySize = 8

        for i in range(subProgramCapacity):
            reader.Position = 4 + i * entrySize
            offset = reader.read_int()
            reader.Position = offset
            self.m_SubPrograms.append(ShaderSubProgram(reader))

    def Export(self, shader):
        shader = re.sub(r"GpuProgramIndex (.+)",
        lambda math: self.m_SubPrograms[int(math.group(1))].Export(),
        shader)

        return shader


class ShaderSubProgram:
    def __init__(self, reader: EndianBinaryReader):
        #LoadGpuProgramFromData
        #201509030 - Unity 5.3
        #201510240 - Unity 5.4
        #201608170 - Unity 5.5
        #201609010 - Unity 5.6, 2017.1 & 2017.2
        #201708220 - Unity 2017.3, Unity 2017.4 & Unity 2018.1
        #201802150 - Unity 2018.2 & Unity 2018.3
        #201806140 - Unity 2019.1~2020.1
        #202012090 - Unity 2021.2

        self.m_Version = reader.read_int()
        self.m_ProgramType = ShaderGpuProgramType(reader.read_int())
        reader.Position += 12

        if self.m_Version >= 201608170:
            reader.Position += 4

        m_KeywordSize = reader.read_int()
        self.m_Keywords = []

        for i in range(m_KeywordSize):
            self.m_Keywords.append(reader.read_aligned_string())

        if 201806140 <= self.m_Version < 202012090:
            m_LocalKeywordsSize = reader.read_int()
            self.m_LocalKeywords = []

            for i in range(m_LocalKeywordsSize):
                self.m_LocalKeywords.append(reader.read_aligned_string())

        self.m_ProgramCode = reader.read_byte_array()
        reader.align_stream()

    def Export(self):
        sb = []

        if len(self.m_Keywords) > 0:
            sb.append("Keywords { ")
            for keyword in self.m_Keywords:
                sb.append("\"{0}\" ".format(keyword))

            sb.append("}\n")

        if hasattr(self, 'm_LocalKeywords') and len(self.m_LocalKeywords) > 0:
            sb.append("Local Keywords { ")
            for keyword in self.m_LocalKeywords:
                sb.append("\"{0}\" ".format(keyword))

            sb.append("}\n")

        sb.append("\"")

        if len(self.m_ProgramCode) > 0:
            if self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramGLLegacy,
                ShaderGpuProgramType.kShaderGpuProgramGLES31AEP,
                ShaderGpuProgramType.kShaderGpuProgramGLES31,
                ShaderGpuProgramType.kShaderGpuProgramGLES3,
                ShaderGpuProgramType.kShaderGpuProgramGLES,
                ShaderGpuProgramType.kShaderGpuProgramGLCore32,
                ShaderGpuProgramType.kShaderGpuProgramGLCore41,
                ShaderGpuProgramType.kShaderGpuProgramGLCore43
            ]:
                sb.append(bytes(self.m_ProgramCode).decode("utf8"))
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM20,
                ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM30,
                ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM20,
                ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM30
            ]:
                sb.append("// shader disassembly not supported on DXBC")
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramDX10Level9Vertex,
                ShaderGpuProgramType.kShaderGpuProgramDX10Level9Pixel,
                ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM40,
                ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM40,
                ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM40,
                ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11HullSM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11DomainSM50
            ]:
                sb.append("// shader disassembly not supported on DXBC")
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramMetalVS,
                ShaderGpuProgramType.kShaderGpuProgramMetalFS
            ]:
                reader = EndianBinaryReader(self.m_ProgramCode, endian = "<")
                fourCC = reader.read_u_int()
                if fourCC == 0xf00dcafe:
                    offset = reader.read_int()
                    reader.Position = offset

                entryName = reader.read_string_to_null()
                buff = reader.read_bytes(int(reader.Length - reader.Position))
                sb.append(bytes(buff).decode("utf8"))
            elif self.m_ProgramType == ShaderGpuProgramType.kShaderGpuProgramSPIRV:
                # TODO SpirVShaderConverter
                pass
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramConsoleVS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleFS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleHS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleDS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleGS
            ]:
                sb.append(bytes(self.m_ProgramCode).decode("utf8"))
            else:
                sb.append("//shader disassembly not supported on {0}".format(self.m_ProgramType))

        sb.append('"')
        return "".join(sb)



from enum import IntEnum

from .NamedObject import NamedObject
from ..export.ShaderConverter import export_shader
from ..enums import ShaderCompilerPlatform, ShaderGpuProgramType, SerializedPropertyType
from ..enums import TextureDimension, PassType


class Shader(NamedObject):
    def export(self):
        return export_shader(self)

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = reader.version
        if version >= (5, 5):  # 5.5 and up
            self.m_ParsedForm = SerializedShader(reader)
            self.platforms = [
                ShaderCompilerPlatform(x) for x in reader.read_u_int_array()
            ]

            if version >= (2019, 3):  # 2019.3 and up
                self.offsets = reader.read_u_int_array_array()[0]
                self.compressedLengths = reader.read_u_int_array_array()[0]
                self.decompressedLengths = reader.read_u_int_array_array()[0]
            else:
                self.offsets = reader.read_u_int_array()
                self.compressedLengths = reader.read_u_int_array()
                self.decompressedLengths = reader.read_u_int_array()
            self.compressedBlob = reader.read_bytes(reader.read_int())
        else:
            self.m_Script = reader.read_bytes(reader.read_int())
            reader.align_stream()
            self.m_PathName = reader.read_aligned_string()
            if version >= (5, 3):  # 5.3 - 5.4
                self.decompressedSize = reader.read_u_int()
                self.m_SubProgramBlob = reader.read_bytes(reader.read_int())


class StructParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_ArraySize = reader.read_int()
        self.m_StructSize = reader.read_int()

        numVectorParams = reader.read_int()
        self.m_VectorParams = [VectorParameter(reader) for _ in range(numVectorParams)]

        numMatrixParams = reader.read_int()
        self.m_MatrixParams = [MatrixParameter(reader) for _ in range(numMatrixParams)]


class SamplerParameter:
    def __init__(self, reader):
        self.sampler = reader.read_u_int()
        self.bindPoint = reader.read_int()


class SerializedTextureProperty:
    def __init__(self, reader):
        self.m_DefaultName = reader.read_aligned_string()
        self.m_TexDim = TextureDimension(reader.read_int())


class SerializedProperty:
    def __init__(self, reader):
        self.m_Name = reader.read_aligned_string()
        self.m_Description = reader.read_aligned_string()
        self.m_Attributes = reader.read_string_array()
        self.m_Type = SerializedPropertyType(reader.read_int())
        self.m_Flags = reader.read_u_int()
        self.m_DefValue = reader.read_float_array(4)
        self.m_DefTexture = SerializedTextureProperty(reader)


class SerializedProperties:
    def __init__(self, reader):
        numProps = reader.read_int()
        self.m_Props = [SerializedProperty(reader) for _ in range(numProps)]


class SerializedShaderFloatValue:
    def __init__(self, reader):
        self.val = reader.read_float()
        self.name = reader.read_aligned_string()


class SerializedShaderRTBlendState:
    def __init__(self, reader):
        self.srcBlend = SerializedShaderFloatValue(reader)
        self.destBlend = SerializedShaderFloatValue(reader)
        self.srcBlendAlpha = SerializedShaderFloatValue(reader)
        self.destBlendAlpha = SerializedShaderFloatValue(reader)
        self.blendOp = SerializedShaderFloatValue(reader)
        self.blendOpAlpha = SerializedShaderFloatValue(reader)
        self.colMask = SerializedShaderFloatValue(reader)


class SerializedStencilOp:
    def __init__(self, reader):
        self.pass_ = SerializedShaderFloatValue(reader)
        self.fail = SerializedShaderFloatValue(reader)
        self.zFail = SerializedShaderFloatValue(reader)
        self.comp = SerializedShaderFloatValue(reader)


class SerializedShaderVectorValue:
    def __init__(self, reader):
        self.x = SerializedShaderFloatValue(reader)
        self.y = SerializedShaderFloatValue(reader)
        self.z = SerializedShaderFloatValue(reader)
        self.w = SerializedShaderFloatValue(reader)
        self.name = reader.read_aligned_string()


class FogMode(IntEnum):
    kFogUnknown = (-1,)
    kFogDisabled = (0,)
    kFogLinear = (1,)
    kFogExp = (2,)
    kFogExp2 = (3,)


class SerializedShaderState:
    def __init__(self, reader):
        version = reader.version

        self.m_Name = reader.read_aligned_string()
        self.rtBlend = [SerializedShaderRTBlendState(reader) for _ in range(8)]
        self.rtSeparateBlend = reader.read_boolean()
        reader.align_stream()
        if version >= (2017, 2):  # 2017.2 and up
            self.zClip = SerializedShaderFloatValue(reader)
        self.zTest = SerializedShaderFloatValue(reader)
        self.zWrite = SerializedShaderFloatValue(reader)
        self.culling = SerializedShaderFloatValue(reader)
        if version >= (2020,):  # 2020.1 and up
            self.conservative = SerializedShaderFloatValue(reader)
        self.offsetFactor = SerializedShaderFloatValue(reader)
        self.offsetUnits = SerializedShaderFloatValue(reader)
        self.alphaToMask = SerializedShaderFloatValue(reader)
        self.stencilOp = SerializedStencilOp(reader)
        self.stencilOpFront = SerializedStencilOp(reader)
        self.stencilOpBack = SerializedStencilOp(reader)
        self.stencilReadMask = SerializedShaderFloatValue(reader)
        self.stencilWriteMask = SerializedShaderFloatValue(reader)
        self.stencilRef = SerializedShaderFloatValue(reader)
        self.fogStart = SerializedShaderFloatValue(reader)
        self.fogEnd = SerializedShaderFloatValue(reader)
        self.fogDensity = SerializedShaderFloatValue(reader)
        self.fogColor = SerializedShaderVectorValue(reader)
        self.fogMode = FogMode(reader.read_int())
        self.gpuProgramID = reader.read_int()
        self.m_Tags = SerializedTagMap(reader)
        self.m_LOD = reader.read_int()
        self.lighting = reader.read_boolean()
        reader.align_stream()


class ShaderBindChannel:
    def __init__(self, reader):
        self.source = reader.read_byte()
        self.target = reader.read_byte()


class ParserBindChannels:
    def __init__(self, reader):
        numChannels = reader.read_int()
        self.m_Channels = [ShaderBindChannel(reader) for _ in range(numChannels)]
        reader.align_stream()
        self.m_SourceMap = reader.read_u_int()


class VectorParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_ArraySize = reader.read_int()
        self.m_Type = reader.read_byte()
        self.m_Dim = reader.read_byte()
        reader.align_stream()


class MatrixParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_ArraySize = reader.read_int()
        self.m_Type = reader.read_byte()
        self.m_RowCount = reader.read_byte()
        reader.align_stream()


class TextureParameter:
    def __init__(self, reader):
        version = reader.version
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_SamplerIndex = reader.read_int()
        if version >= (2017, 3):  # 2017.3 and up
            self.m_MultiSampled = reader.read_boolean()
        self.m_Dim = reader.read_byte()
        reader.align_stream()


class BufferBinding:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        if reader.version >= (2020,):  # 2020.1 and up
            m_ArraySize = reader.read_int()


class ConstantBuffer:
    def __init__(self, reader):
        version = reader.version

        self.m_NameIndex = reader.read_int()

        numMatrixParams = reader.read_int()
        self.m_MatrixParams = [MatrixParameter(reader) for _ in range(numMatrixParams)]

        numVectorParams = reader.read_int()
        self.m_VectorParams = [VectorParameter(reader) for _ in range(numVectorParams)]
        if version >= (2017, 3):  # 2017.3 and up
            numStructParams = reader.read_int()
            self.m_StructParams = [
                StructParameter(reader) for _ in range(numStructParams)
            ]
        self.m_Size = reader.read_int()

        if version >= (2021, 1, 4) or (version[0] == 2020 and version >= (2020, 3, 2)):
            self.m_IsPartialCB = reader.read_boolean()
            reader.align_stream()


class UAVParameter:
    def __init__(self, reader):
        self.m_NameIndex = reader.read_int()
        self.m_Index = reader.read_int()
        self.m_OriginalIndex = reader.read_int()


class SerializedProgramParameters:
    def __init__(self, reader):
        numVectorParams = reader.read_int()
        self.m_VectorParams = [VectorParameter(reader) for _ in range(numVectorParams)]

        numMatrixParams = reader.read_int()
        self.m_MatrixParams = [MatrixParameter(reader) for _ in range(numMatrixParams)]

        numTextureParams = reader.read_int()
        self.m_TextureParams = [
            TextureParameter(reader) for _ in range(numTextureParams)
        ]

        numBufferParams = reader.read_int()
        self.m_BufferParams = [BufferBinding(reader) for _ in range(numBufferParams)]

        numConstantBuffers = reader.read_int()
        self.m_ConstantBuffers = [
            ConstantBuffer(reader) for _ in range(numConstantBuffers)
        ]

        numConstantBufferBindings = reader.read_int()
        self.m_ConstantBufferBindings = [
            BufferBinding(reader) for _ in range(numConstantBufferBindings)
        ]

        numUAVParams = reader.read_int()
        self.m_UAVParams = [UAVParameter(reader) for _ in range(numUAVParams)]

        numSamplers = reader.read_int()
        self.m_Samplers = [SamplerParameter(reader) for _ in range(numSamplers)]


class SerializedSubProgram:
    def __init__(self, reader):
        version = reader.version

        self.m_BlobIndex = reader.read_u_int()
        self.m_Channels = ParserBindChannels(reader)

        if (2019,) <= version < (2021, 2):  # 2019 ~2021.1
            self.m_GlobalKeywordIndices = reader.read_u_short_array()
            reader.align_stream()
            self.m_LocalKeywordIndices = reader.read_u_short_array()
            reader.align_stream()
        else:
            self.m_KeywordIndices = reader.read_u_short_array()
            if version >= (2017,):  # 2017 and up
                reader.align_stream()

        self.m_ShaderHardwareTier = reader.read_byte()
        self.m_GpuProgramType = ShaderGpuProgramType(reader.read_byte())
        reader.align_stream()

        if version >= (2021, 1, 4) or (version[0] == 2020 and version >= (2020, 3, 2)):
            self.m_Parameters = SerializedProgramParameters(reader)
        else:
            numVectorParams = reader.read_int()
            self.m_VectorParams = [
                VectorParameter(reader) for _ in range(numVectorParams)
            ]

            numMatrixParams = reader.read_int()
            self.m_MatrixParams = [
                MatrixParameter(reader) for _ in range(numMatrixParams)
            ]

            numTextureParams = reader.read_int()
            self.m_TextureParams = [
                TextureParameter(reader) for _ in range(numTextureParams)
            ]

            numBufferParams = reader.read_int()
            self.m_BufferParams = [
                BufferBinding(reader) for _ in range(numBufferParams)
            ]

            numConstantBuffers = reader.read_int()
            self.m_ConstantBuffers = [
                ConstantBuffer(reader) for _ in range(numConstantBuffers)
            ]

            numConstantBufferBindings = reader.read_int()
            self.m_ConstantBufferBindings = [
                BufferBinding(reader) for _ in range(numConstantBufferBindings)
            ]

            numUAVParams = reader.read_int()
            self.m_UAVParams = [UAVParameter(reader) for _ in range(numUAVParams)]

            if version >= (2017,):  # 2017 and up
                numSamplers = reader.read_int()
                self.m_Samples = [SamplerParameter(reader) for _ in range(numSamplers)]

        if version >= (2017, 2):  # 2017.2 and up
            if version >= (2021,):
                self.m_ShaderRequirements = reader.read_long()
            else:
                self.m_ShaderRequirements = reader.read_int()


class SerializedProgram:
    def __init__(self, reader):
        version = reader.version

        numSubPrograms = reader.read_int()
        self.m_SubPrograms = [
            SerializedSubProgram(reader) for _ in range(numSubPrograms)
        ]

        if version >= (2021, 1, 4) or (version[0] == 2020 and version >= (2020, 3, 2)):
            self.m_CommonParameters = SerializedProgramParameters(reader)


class SerializedPass:
    def __init__(self, reader):
        version = reader.version

        if version >= (2020, 2):  # 2020.2 and up
            numEditorDataHash = reader.read_int()
            m_EditorDataHash = [
                reader.read_bytes(16)  # Hash128(reader)
                for _ in range(numEditorDataHash)
            ]
            reader.align_stream()
            m_Platforms = reader.read_byte_array()
            reader.align_stream()
            if version < (2021, 2):
                m_LocalKeywordMask = reader.read_u_short_array()
                reader.align_stream()
                m_GlobalKeywordMask = reader.read_u_short_array()
                reader.align_stream()

        numIndices = reader.read_int()
        self.m_NameIndices = {}
        for _ in range(numIndices):
            key = reader.read_aligned_string()
            self.m_NameIndices[key] = reader.read_int()
        self.m_Type = PassType(reader.read_int())
        self.m_State = SerializedShaderState(reader)
        self.m_ProgramMask = reader.read_u_int()
        self.progVertex = SerializedProgram(reader)
        self.progFragment = SerializedProgram(reader)
        self.progGeometry = SerializedProgram(reader)
        self.progHull = SerializedProgram(reader)
        self.progDomain = SerializedProgram(reader)
        if version >= (2019, 3):  # 2019.3 and up
            self.progRayTracing = SerializedProgram(reader)
        self.m_HasInstancingVariant = reader.read_boolean()
        if version >= (2018,):  # 2018 and up
            self.m_HasProceduralInstancingVariant = reader.read_boolean()
        reader.align_stream()
        self.m_UseName = reader.read_aligned_string()
        self.m_Name = reader.read_aligned_string()
        self.m_TextureName = reader.read_aligned_string()
        self.m_Tags = SerializedTagMap(reader)
        if version >= (2021, 2):
            m_SerializedKeywordStateMask = reader.read_u_short_array()
            reader.align_stream()


class SerializedTagMap:
    def __init__(self, reader):
        numTags = reader.read_int()
        self.tags = {}
        for _ in range(numTags):
            key = reader.read_aligned_string()
            self.tags[key] = reader.read_aligned_string()


class SerializedSubShader:
    def __init__(self, reader):
        numPasses = reader.read_int()
        self.m_Passes = [SerializedPass(reader) for _ in range(numPasses)]
        self.m_Tags = SerializedTagMap(reader)
        self.m_LOD = reader.read_int()


class SerializedShaderDependency:
    def __init__(self, reader):
        self.from_ = reader.read_aligned_string()
        self.to = reader.read_aligned_string()


class SerializedCustomEditorForRenderPipeline:
    def __init__(self, reader):
        self.customEditorName = reader.read_aligned_string()
        self.renderPipelineType = reader.read_aligned_string()


class SerializedShader:
    def __init__(self, reader):
        version = reader.version

        self.m_PropInfo = SerializedProperties(reader)
        numSubShaders = reader.read_int()
        self.m_SubShaders = [SerializedSubShader(reader) for _ in range(numSubShaders)]

        if version >= (2021, 2):
            self.m_KeywordNames = reader.read_string_array()
            self.m_KeywordFlags = reader.read_bytes(reader.read_int())
            reader.align_stream()

        self.m_Name = reader.read_aligned_string()
        self.m_CustomEditorName = reader.read_aligned_string()
        self.m_FallbackName = reader.read_aligned_string()
        numDependencies = reader.read_int()
        self.m_Dependencies = [
            SerializedShaderDependency(reader) for _ in range(numDependencies)
        ]

        if version >= (2021,):
            m_CustomEditorForRenderPipelinesSize = reader.read_int()
            self.m_CustomEditorForRenderPipelines = [
                SerializedCustomEditorForRenderPipeline(reader)
                for _ in range(m_CustomEditorForRenderPipelinesSize)
            ]

        self.m_DisableNoSubshadersMessage = reader.read_boolean()
        reader.align_stream()



from ..enums import ClassIDType
from . import SerializedFile
from .. import classes
from ..streams import EndianBinaryReader, EndianBinaryWriter
from ..helpers import TypeTreeHelper


class ObjectReader:
    byte_start: int
    byte_size: int
    type_id: int
    class_id: int
    type: ClassIDType
    path_id: int
    # serialized_type: SerializedType
    _read_until: int

    # saves where the parser stopped
    # in case that not all data is read
    # and the obj.data is changed, the unknown data can be added again

    def __init__(self, assets_file, reader: EndianBinaryReader):
        self.assets_file = assets_file
        self.reader = reader
        self.data = b""
        self.version = assets_file.version
        self.version2 = assets_file.header.version
        self.platform = assets_file.target_platform
        self.build_type = assets_file.build_type

        header = assets_file.header
        types = assets_file.types

        if assets_file.big_id_enabled:
            self.path_id = reader.read_long()
        elif header.version < 14:
            self.path_id = reader.read_int()
        else:
            reader.align_stream()
            self.path_id = reader.read_long()

        if header.version >= 22:
            self.byte_start_offset = (self.reader.real_offset(), 8)
            self.byte_start = reader.read_long()
        else:
            self.byte_start_offset = (self.reader.real_offset(), 4)
            self.byte_start = reader.read_u_int()

        self.byte_start += header.data_offset
        self.byte_header_offset = header.data_offset
        self.byte_base_offset = self.reader.BaseOffset

        self.byte_size_offset = (self.reader.real_offset(), 4)
        self.byte_size = reader.read_u_int()

        self.type_id = reader.read_int()
        if header.version < 16:
            self.class_id = reader.read_u_short()
            if types:
                self.serialized_type = (
                    x for x in types if x.class_id == self.type_id).__next__()
            else:
                self.serialized_type = SerializedFile.SerializedType(
                    reader, self.assets_file)
        else:
            typ = types[self.type_id]
            self.serialized_type = typ
            self.class_id = typ.class_id

        self.type = ClassIDType(self.class_id)

        if header.version < 11:
            self.is_destroyed = reader.read_u_short()

        if 11 <= header.version < 17:
            script_type_index = reader.read_short()
            if self.serialized_type:
                self.serialized_type.script_type_index = script_type_index

        if header.version == 15 or header.version == 16:
            self.stripped = reader.read_byte()

    def write(self, header, writer: EndianBinaryWriter, data_writer: EndianBinaryWriter):
        if self.assets_file.big_id_enabled:
            writer.write_long(self.path_id)
        elif header.version < 14:
            writer.write_int(self.path_id)
        else:
            writer.align_stream()
            writer.write_long(self.path_id)

        if self.data:
            data = self.data
            # in some cases the parser doesn't read all of the object data
            # games might still require the missing data
            # so following code appends the missing data back to edited objects
            # -> following solution has let to some problems, so it will be removed for now
            # if self.type != ClassIDType.MonoBehaviour:
            #     end_pos = self.byte_start + self.byte_size
            #     if self._read_until and self._read_until != end_pos:
            #         self.reader.Position = self._read_until
            #         data += self.reader.read_bytes(end_pos - self._read_until)
        else:
            self.reset()
            data = self.reader.read(self.byte_size)

        if header.version >= 22:
            writer.write_long(data_writer.Position)
        else:
            writer.write_u_int(data_writer.Position)

        writer.write_u_int(len(data))
        data_writer.write(data)

        writer.write_int(self.type_id)

        if header.version < 16:
            writer.write_u_short(self.class_id)

        if header.version < 11:
            writer.write_u_short(self.is_destroyed)

        if 11 <= header.version < 17:
            writer.write_short(self.serialized_type.script_type_index)

        if header.version == 15 or header.version == 16:
            writer.write_byte(self.stripped)

    def set_raw_data(self, data):
        self.data = data
        self.assets_file.mark_changed()

    @property
    def container(self):
        return (
            self.assets_file._container[self.path_id]
            if self.path_id in self.assets_file._container
            else None
        )

    @property
    def Position(self):
        return self.reader.Position

    @Position.setter
    def Position(self, pos):
        self.reader.Position = pos

    def reset(self):
        self.reader.Position = self.byte_start

    def read(self):
        try:
            obj = getattr(classes, self.type.name, classes.Object)(self)
        except Exception as e:
            raise e
        # TODO: only specific exceptions here?
        except:
            # HACK: in case the parsing via the class fails this solution
            #       uses the type tree to set the variables and then changes the class
            obj = classes.Object(self)
            obj.__class__ = getattr(classes, self.type.name, classes.Object)
            for key, val in obj.__dict__.items():
                if " " in key:
                    obj.__dict__[key.replace(" ", "_")] = val
                    delattr(obj, key)
        self._read_until = self.reader.Position
        return obj

    def get(self, key, default=None):
        return getattr(self, key, default)

    def __getattr__(self, item: str):
        if hasattr(self.reader, item):
            return getattr(self.reader, item)

    def __repr__(self):
        return "<%s %s>" % (self.__class__.__name__, self.type.name)

    ###################################################
    #
    #           Typetree Stuff
    #
    ###################################################

    def dump_typetree(self, nodes: list = None) -> str:
        self.reset()
        sb = []
        if nodes:
            TypeTreeHelper.read_typetree_str(sb, nodes, self)
        elif getattr(self.serialized_type, "nodes", None):
            TypeTreeHelper.read_typetree_str(
                sb, self.serialized_type.nodes, self)
            return "".join(sb)
        return ""

    def dump_typetree_structure(self) -> str:
        if getattr(self.serialized_type, "nodes", None):
            return TypeTreeHelper.dump_typetree(self.serialized_type.nodes)
        return ""

    def read_typetree(self, nodes: list = None) -> dict:
        self.reset()
        tree = {}
        if nodes:
            tree = TypeTreeHelper.read_typetree(nodes, self)
        elif self.serialized_type.nodes:
            tree = TypeTreeHelper.read_typetree(
                self.serialized_type.nodes, self)
        return tree

    def save_typetree(self, tree: dict, nodes: list = None, writer: EndianBinaryWriter = None):
        if not writer:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        if nodes:
            TypeTreeHelper.write_typetree(tree, nodes, writer)
        elif self.serialized_type.nodes:
            TypeTreeHelper.write_typetree(
                tree, self.serialized_type.nodes, writer)
        else:
            raise ValueError("There are no TypeTree nodes for this object.")
        data = writer.bytes
        self.set_raw_data(data)
        return data

    def get_raw_data(self) -> bytes:
        pos = self.Position
        self.reset()
        ret = self.reader.read_bytes(self.byte_size)
        self.Position = pos
        return ret

    def set_raw_data(self, data):
        self.data = data
        if self.assets_file:
            self.assets_file.mark_changed()




from UnityPy.classes import Mesh


def export_mesh(m_Mesh: Mesh, format="obj") -> str:
    if format == "obj":
        return export_mesh_obj(m_Mesh)
    raise NotImplementedError(f"Export format {format} not implemented")


def export_mesh_obj(m_Mesh, material_names: list = None):
    if m_Mesh.m_VertexCount <= 0:
        return False

    sb = [f"g {m_Mesh.name}\r\n"]
    if material_names:
        sb.append(f"mtllib {m_Mesh.name}.mtl\r\n")
    # region Vertices
    if not m_Mesh.m_Vertices:
        return False

    c = 3
    if len(m_Mesh.m_Vertices) == m_Mesh.m_VertexCount * 4:
        c = 4

    for v in range(int(m_Mesh.m_VertexCount)):
        sb.append(
            "v {0:.7G} {1:.7G} {2:.7G}\r\n".format(
                -m_Mesh.m_Vertices[v * c],
                m_Mesh.m_Vertices[v * c + 1],
                m_Mesh.m_Vertices[v * c + 2],
            ).replace("nan", "0")
        )
    # endregion

    # region UV
    if m_Mesh.m_UV0:
        if len(m_Mesh.m_UV0) == m_Mesh.m_VertexCount * 2:
            c = 2
        elif len(m_Mesh.m_UV0) == m_Mesh.m_VertexCount * 3:
            c = 3

        for v in range(int(m_Mesh.m_VertexCount)):
            sb.append(
                "vt {0:.7G} {1:.7G}\r\n".format(
                    m_Mesh.m_UV0[v * c], m_Mesh.m_UV0[v * c + 1]
                ).replace("nan", "0")
            )
    # endregion

    # region Normals
    if m_Mesh.m_Normals:
        if len(m_Mesh.m_Normals) == m_Mesh.m_VertexCount * 3:
            c = 3
        elif len(m_Mesh.m_Normals) == m_Mesh.m_VertexCount * 4:
            c = 4

        for v in range(int(m_Mesh.m_VertexCount)):
            sb.append(
                "vn {0:.7G} {1:.7G} {2:.7G}\r\n".format(
                    -m_Mesh.m_Normals[v * c],
                    m_Mesh.m_Normals[v * c + 1],
                    m_Mesh.m_Normals[v * c + 2],
                ).replace("nan", "0")
            )
    # endregion

    # region Face
    sum = 0
    for i in range(len(m_Mesh.m_SubMeshes)):
        sb.append(f"g {m_Mesh.name}_{i}\r\n")
        if material_names and i < len(material_names) and material_names[i]:
            sb.append(f"usemtl {material_names[i]}\r\n")
        indexCount = m_Mesh.m_SubMeshes[i].indexCount
        end = sum + indexCount // 3
        for f in range(sum, end):
            sb.append(
                "f {0}/{0}/{0} {1}/{1}/{1} {2}/{2}/{2}\r\n".format(
                    m_Mesh.m_Indices[f * 3 + 2] + 1,
                    m_Mesh.m_Indices[f * 3 + 1] + 1,
                    m_Mesh.m_Indices[f * 3] + 1,
                )
            )
        sum = end
    # endregion
    return "".join(sb)



from .Audio import AudioType, AudioCompressionFormat, AUDIO_TYPE_EXTEMSION
from .BuildTarget import BuildTarget
from .ClassIDType import ClassIDType
from .FileType import FileType
from .TextureFormat import TextureFormat
from .SpriteMeshType import SpriteMeshType
from .GfxPrimitiveType import GfxPrimitiveType
from .CommonString import CommonString
from .ShaderCompilerPlatform import ShaderCompilerPlatform
from .ShaderGpuProgramType import ShaderGpuProgramType
from .SerializedPropertyType import SerializedPropertyType
from .SpritePackingMode import SpritePackingMode
from .SpritePackingRotation import SpritePackingRotation
from .TextureDimension import TextureDimension
from .PassType import PassType



import os
from typing import Union, List
from .CompressionHelper import BROTLI_MAGIC, GZIP_MAGIC
from ..enums import FileType
from ..streams import EndianBinaryReader


def file_name_without_extension(file_name: str) -> str:
    return os.path.join(
        os.path.dirname(file_name), os.path.splitext(os.path.basename(file_name))[0]
    )


def list_all_files(directory: str) -> List[str]:
    return [
        val
        for sublist in [
            [os.path.join(dir_path, filename) for filename in filenames]
            for (dir_path, dirn_ames, filenames) in os.walk(directory)
            if ".git" not in dir_path
        ]
        for val in sublist
    ]


def find_all_files(directory: str, search_str: str) -> List[str]:
    return [
        val
        for sublist in [
            [
                os.path.join(dir_path, filename)
                for filename in filenames
                if search_str in filename
            ]
            for (dir_path, dirn_ames, filenames) in os.walk(directory)
            if ".git" not in dir_path
        ]
        for val in sublist
    ]


def check_file_type(input_) -> Union[FileType, EndianBinaryReader]:
    if isinstance(input_, str) and os.path.isfile(input_):
        reader = EndianBinaryReader(open(input_, "rb"))
    elif isinstance(input_, EndianBinaryReader):
        reader = input_
    else:
        try:
            reader = EndianBinaryReader(input_)
        except:
            return None, None

    if reader.Length < 20:
        return FileType.ResourceFile, reader

    signature = reader.read_string_to_null(20)
    reader.Position = 0
    if signature in [
        "UnityWeb",
        "UnityRaw",
        "\xFA\xFA\xFA\xFA\xFA\xFA\xFA\xFA",
        "UnityFS",
    ]:
        return FileType.BundleFile, reader
    elif signature == "UnityWebData1.0":
        return FileType.WebFile, reader
    elif signature == "PK\x03\x04":
        return FileType.ZIP, reader
    else:
        if reader.Length < 128:
            return FileType.ResourceFile, reader

        magic = bytes(reader.read_bytes(2))
        reader.Position = 0
        if GZIP_MAGIC == magic:
            return FileType.WebFile, reader
        reader.Position = 0x20
        magic = bytes(reader.read_bytes(6))
        reader.Position = 0
        if BROTLI_MAGIC == magic:
            return FileType.WebFile, reader

        # check if AssetsFile
        old_endian = reader.endian
        # read as if assetsfile and check version
        # ReadHeader
        reader.Position = 0
        metadata_size = reader.read_u_int()
        file_size = reader.read_u_int()
        version = reader.read_u_int()
        data_offset = reader.read_u_int()

        if version >= 22:
            endian = ">" if reader.read_boolean() else "<"
            reserved = reader.read_bytes(3)
            metadata_size = reader.read_u_int()
            file_size = reader.read_long()
            data_offset = reader.read_long()
            unknown = reader.read_long()  # unknown

        # reset
        reader.endian = old_endian
        reader.Position = 0
        # check info
        if any(
            (
                version < 0,
                version > 100,
                *[
                    x < 0 or x > reader.Length
                    for x in [file_size, metadata_size, version, data_offset]
                ],
                file_size < metadata_size,
                file_size < data_offset,
            )
        ):
            return FileType.ResourceFile, reader
        else:
            return FileType.AssetsFile, reader



from .NamedObject import NamedObject


class MonoScript(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        if version >= (3, 4):  # 3.4 and up
            self.m_ExecutionOrder = reader.read_int()
        if version < (5,):  # 5.0 down
            self.m_PropertiesHash = reader.read_u_int()
        else:
            self.m_PropertiesHash = reader.read_bytes(16)
        if version < (3,):  # 3.0 down
            self.m_PathName = reader.read_aligned_string()

        self.m_ClassName = reader.read_aligned_string()
        if version >= (3,):  # 3.0 and up
            self.m_Namespace = reader.read_aligned_string()

        self.m_AssemblyName = reader.read_aligned_string()
        if version < (2018, 2):  # 2018.2 down
            self.m_IsEditorScript = reader.read_boolean()



import os
import re
import sys

from . import File, ObjectReader
from ..enums import BuildTarget, ClassIDType, CommonString
from ..streams import EndianBinaryReader, EndianBinaryWriter
from ..helpers.TypeTreeHelper import TypeTreeNode

RECURSION_LIMIT = sys.getrecursionlimit()


class SerializedFileHeader:
    metadata_size: int
    file_size: int
    version: int
    data_offset: int
    endian: bytes
    reserved: bytes

    def __init__(self, reader: EndianBinaryReader):
        self.metadata_size = reader.read_u_int()
        self.file_size = reader.read_u_int()
        self.version = reader.read_u_int()
        self.data_offset = reader.read_u_int()


class LocalSerializedObjectIdentifier:  # script type
    local_serialized_file_index: int
    local_identifier_in_file: int

    def __init__(self, header: SerializedFileHeader, reader: EndianBinaryReader):
        self.local_serialized_file_index = reader.read_int()
        if header.version < 14:
            self.local_identifier_in_file = reader.read_int()
        else:
            reader.align_stream()
            self.local_identifier_in_file = reader.read_long()

    def write(self, header: SerializedFileHeader, writer: EndianBinaryWriter):
        writer.write_int(self.local_serialized_file_index)
        if header.version < 14:
            writer.write_int(self.local_identifier_in_file)
        else:
            writer.align_stream()
            writer.write_long(self.local_identifier_in_file)


class FileIdentifier:  # external
    guid: bytes
    type: int
    # enum { kNonAssetType = 0, kDeprecatedCachedAssetType = 1, kSerializedAssetType = 2, kMetaAssetType = 3 };
    path: str

    @property
    def name(self):
        return os.path.basename(self.path)

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.path})>"

    def __init__(self, header: SerializedFileHeader, reader: EndianBinaryReader):
        if header.version >= 6:
            self.temp_empty = reader.read_string_to_null()
        if header.version >= 5:
            self.guid = reader.read_bytes(16)
            self.type = reader.read_int()
        self.path = reader.read_string_to_null()

    def write(self, header: SerializedFileHeader, writer: EndianBinaryWriter):
        if header.version >= 6:
            writer.write_string_to_null(self.temp_empty)
        if header.version >= 5:
            writer.write_bytes(self.guid)
            writer.write_int(self.type)
        writer.write_string_to_null(self.path)


class BuildType:
    build_type: str

    def __init__(self, build_type):
        self.build_type = build_type

    @property
    def is_alpha(self):
        return self.build_type == "a"

    @property
    def is_path(self):
        return self.build_type == "p"


class SerializedType:
    class_id: int
    is_stripped_type: bool
    script_type_index = -1
    nodes: list = []  # TypeTreeNode
    script_id: bytes  # Hash128
    old_type_hash: bytes  # Hash128}

    def __init__(self, reader, serialized_file):
        version = serialized_file.header.version
        self.class_id = reader.read_int()

        if version >= 16:
            self.is_stripped_type = reader.read_boolean()

        if version >= 17:
            self.script_type_index = reader.read_short()

        if version >= 13:
            if (version < 16 and self.class_id < 0) or (
                version >= 16 and self.class_id == 114
            ):
                self.script_id = reader.read_bytes(16)  # Hash128
            self.old_type_hash = reader.read_bytes(16)  # Hash128

        if serialized_file._enable_type_tree:
            type_tree = []
            if version >= 12 or version == 10:
                self.string_data = serialized_file.read_type_tree_blob(type_tree)
            else:
                serialized_file.read_type_tree(type_tree)

            if version >= 21:
                self.type_dependencies = reader.read_int_array()

            self.nodes = type_tree

    def write(self, serialized_file, writer):
        version = serialized_file.header.version
        writer.write_int(self.class_id)

        if version >= 16:
            writer.write_boolean(self.is_stripped_type)

        if version >= 17:
            writer.write_short(self.script_type_index)

        if version >= 13:
            if (version < 16 and self.class_id < 0) or (
                version >= 16 and self.class_id == 114
            ):
                writer.write_bytes(self.script_id)  # Hash128
            writer.write_bytes(self.old_type_hash)  # Hash128

        if serialized_file._enable_type_tree:
            if version >= 12 or version == 10:
                serialized_file.save_type_tree5(self.nodes, writer, self.string_data)
            else:
                serialized_file.save_type_tree(self.nodes, writer)


class SerializedFile(File.File):
    reader: EndianBinaryReader
    is_changed: bool
    unity_version: str
    version: tuple
    build_type: BuildType
    target_platform: BuildTarget
    types: list
    script_types: list
    externals: list
    _container: dict
    objects: dict
    container_: dict
    _cache: dict
    header: SerializedFileHeader

    @property
    def files(self):
        if self.objects:
            return self.objects
        return {}

    @files.setter
    def files(self, value):
        self.objects = value

    def __init__(self, reader: EndianBinaryReader, parent=None, name=None):
        super().__init__(parent=parent, name=name)
        self.reader = reader

        self.unity_version = "2.5.0f5"
        self.version = (0, 0, 0, 0)
        self.build_type = BuildType("")
        self.target_platform = BuildTarget.UnknownPlatform
        self._enable_type_tree = True
        self.types = []
        self.script_types = []
        self.externals = []
        self._container = {}

        self.objects = {}
        self.container_ = {}
        # used to speed up mass asset extraction
        # some assets refer to each other, so by keeping the result
        # of specific assets cached the extraction can be speed up by a lot.
        # used by: Sprite (Texture2D (with alpha) cached),
        self._cache = {}
        self.unknown = 0

        # ReadHeader
        header = SerializedFileHeader(reader)
        self.header = header

        if header.version >= 9:
            header.endian = ">" if reader.read_boolean() else "<"
            header.reserved = reader.read_bytes(3)
            if header.version >= 22:
                header.metadata_size = reader.read_u_int()
                header.file_size = reader.read_long()
                header.data_offset = reader.read_long()
                self.unknown = reader.read_long()  # unknown
        else:
            reader.Position = header.file_size - header.metadata_size
            header.endian = ">" if reader.read_boolean() else "<"

        reader.endian = header.endian

        if header.version >= 7:
            unity_version = reader.read_string_to_null()
            self.set_version(unity_version)

        if header.version >= 8:
            self._m_target_platform = reader.read_int()
            self.target_platform = BuildTarget(self._m_target_platform)

        if header.version >= 13:
            self._enable_type_tree = reader.read_boolean()

        # ReadTypes
        type_count = reader.read_int()
        self.types = [SerializedType(reader, self) for _ in range(type_count)]

        self.big_id_enabled = 0
        if 7 <= header.version < 14:
            self.big_id_enabled = reader.read_int()

        # ReadObjects
        object_count = reader.read_int()
        self.objects = {}
        for _ in range(object_count):
            obj = ObjectReader.ObjectReader(self, reader)
            self.objects[obj.path_id] = obj

        # Read Scripts
        if header.version >= 11:
            script_count = reader.read_int()
            self.script_types = [
                LocalSerializedObjectIdentifier(header, reader)
                for _ in range(script_count)
            ]

        # Read Externals
        externals_count = reader.read_int()
        self.externals = [
            FileIdentifier(header, reader) for _ in range(externals_count)
        ]

        if header.version >= 20:
            ref_type_count = reader.read_int()
            self.ref_types = [
                SerializedType(reader, self) for _ in range(ref_type_count)
            ]

        if header.version >= 5:
            self.userInformation = reader.read_string_to_null()

        # read the asset_bundles to get the containers
        for obj in self.objects.values():
            if obj.type == ClassIDType.AssetBundle:
                data = obj.read()
                for container, asset_info in data.m_Container.items():
                    asset = asset_info.asset
                    self.container_[container] = asset
                    if hasattr(asset, "path_id"):
                        self._container[asset.path_id] = container
        # if environment is not None:
        #    environment.container = {**environment.container, **self.container}

    @property
    def container(self):
        return self.container_

    def set_version(self, string_version):
        self.unity_version = string_version
        if string_version == "0.0.0":
            # weird case, but apparently can happen?
            # check "cant read Texture2D by 2020.3.13 f1 AssetBundle #77" for details
            string_version = self.parent.version_engine
        build_type = re.findall(r"([^\d.])", string_version)
        self.build_type = BuildType(build_type[0] if build_type else "")
        version_split = re.split(r"\D", string_version)
        self.version = tuple(int(x) for x in version_split)

    def read_type_tree(self, type_tree):
        level_stack = [[0, 1]]
        while level_stack:
            level, count = level_stack[-1]
            if count == 1:
                level_stack.pop()
            else:
                level_stack[-1][1] -= 1

            type_tree_node = TypeTreeNode()
            type_tree.append(type_tree_node)
            type_tree_node.level = level
            type_tree_node.type = self.reader.read_string_to_null()
            type_tree_node.name = self.reader.read_string_to_null()
            type_tree_node.byte_size = self.reader.read_int()
            if self.header.version == 2:
                type_tree_node.variable_count = self.reader.read_int()

            if self.header.version != 3:
                type_tree_node.index = self.reader.read_int()

            type_tree_node.is_array = self.reader.read_int()
            type_tree_node.version = self.reader.read_int()
            if self.header.version != 3:
                type_tree_node.meta_flag = self.reader.read_int()

            children_count = self.reader.read_int()
            if children_count:
                level_stack.append([level + 1, children_count])
        return type_tree

    def read_type_tree_blob(self, type_tree):
        reader = self.reader
        number_of_nodes = self.reader.read_int()
        string_buffer_size = self.reader.read_int()

        for _ in range(number_of_nodes):
            node = TypeTreeNode()
            type_tree.append(node)
            node.version = reader.read_u_short()
            node.level = reader.read_byte()
            node.is_array = reader.read_boolean()
            node.type_str_offset = reader.read_u_int()
            node.name_str_offset = reader.read_u_int()
            node.byte_size = reader.read_int()
            node.index = reader.read_int()
            node.meta_flag = reader.read_int()

            if self.header.version >= 19:
                node.ref_type_hash = reader.read_u_long()

        string_buffer_reader = EndianBinaryReader(
            reader.read(string_buffer_size), reader.endian
        )
        for node in type_tree:
            node.type = read_string(string_buffer_reader, node.type_str_offset)
            node.name = read_string(string_buffer_reader, node.name_str_offset)

        return string_buffer_reader.bytes

    def get_writeable_cab(self, name: str = "CAB-UnityPy_Mod.resS"):
        """
        Creates a new cab file in the bundle that contains the given data.
        This is usefull for asset types that use resource files.
        """
        if not isinstance(
            self.parent, (File.BundleFile.BundleFile, File.WebFile.WebFile)
        ):
            return None
        
        cab = self.parent.get_writeable_cab(name)
        cab.path = f"archive:/{self.name}/{name}"
        if not any(cab.path == x.path for x in self.externals):
            # register as external
            class FileIdentifierFake:
                pass

            file_identifier = FileIdentifierFake()
            file_identifier.__class__ = FileIdentifier
            file_identifier.temp_empty = ""
            import uuid

            file_identifier.guid = uuid.uuid1().urn[-16:].encode("ascii")
            file_identifier.path = cab.path
            file_identifier.type = 0
            self.externals.append(file_identifier)

        return cab

    def save(self, packer: str = None) -> bytes:
        # 1. header -> has to be delayed until the very end
        # 2. data -> types, objects, scripts, ...

        # so write the data first
        header = self.header
        meta_writer = EndianBinaryWriter(endian=header.endian)
        data_writer = EndianBinaryWriter(endian=header.endian)
        if header.version >= 7:
            meta_writer.write_string_to_null(self.unity_version)

        if header.version >= 8:
            meta_writer.write_int(self._m_target_platform)

        if header.version >= 13:
            meta_writer.write_boolean(self._enable_type_tree)

        # ReadTypes
        meta_writer.write_int(len(self.types))
        for typ in self.types:
            typ.write(self, meta_writer)

        if 7 <= header.version < 14:
            meta_writer.write_int(self.big_id_enabled)

        # ReadObjects
        meta_writer.write_int(len(self.objects))
        for obj in self.objects.values():
            obj.write(header, meta_writer, data_writer)
            data_writer.align_stream(8)

        # Read Scripts
        if header.version >= 11:
            meta_writer.write_int(len(self.script_types))
            for script_type in self.script_types:
                script_type.write(header, meta_writer)

        # Read Externals
        meta_writer.write_int(len(self.externals))
        for external in self.externals:
            external.write(header, meta_writer)

        if header.version >= 20:
            meta_writer.write_int(len(self.ref_types))
            for ref_type in self.ref_types:
                ref_type.write(self, meta_writer)

        if header.version >= 5:
            meta_writer.write_string_to_null(self.userInformation)

        # prepare header
        writer = EndianBinaryWriter()
        header_size = 16  # 4*4
        metadata_size = meta_writer.Length
        data_size = data_writer.Length
        if header.version >= 9:
            # 1 bool + 3 reserved + extra header 4 + 3*8
            header_size += 4 if header.version < 22 else 4 + 28
            data_offset = header_size + metadata_size
            # align data_offset
            data_offset += (16 - data_offset % 16) % 16
            file_size = data_offset + data_size
            if header.version < 22:
                writer.write_u_int(metadata_size)
                writer.write_u_int(file_size)
                writer.write_u_int(header.version)
                # reader.Position = header.file_size - header.metadata_size
                # so data follows right after this header -> after 32
                writer.write_u_int(data_offset)
                writer.write_boolean(">" == header.endian)
                writer.write_bytes(header.reserved)
            else:
                # old header
                writer.write_u_int(0)
                writer.write_u_int(0)
                writer.write_u_int(header.version)
                writer.write_u_int(0)
                writer.write_boolean(">" == header.endian)
                writer.write_bytes(header.reserved)
                writer.write_u_int(metadata_size)
                writer.write_long(file_size)
                writer.write_long(data_offset)
                writer.write_long(self.unknown)

            writer.write_bytes(meta_writer.bytes)
            writer.align_stream(16)
            writer.write_bytes(data_writer.bytes)

        else:
            metadata_size += 1  # endian boolean
            file_size = header_size + metadata_size + data_size
            writer.write_u_int(metadata_size)
            writer.write_u_int(file_size)
            writer.write_u_int(header.version)
            # reader.Position = header.file_size - header.metadata_size
            # so data follows right after this header -> after 32
            writer.write_u_int(32)
            writer.write_bytes(data_writer.bytes)
            writer.write_boolean(">" == header.endian)
            writer.write_bytes(meta_writer.bytes)

        return writer.bytes

    def save_serialized_type(
        self,
        typ: SerializedType,
        header: SerializedFileHeader,
        writer: EndianBinaryWriter,
    ):
        writer.write_int(typ.class_id)

        if header.version >= 16:
            writer.write_boolean(typ.is_stripped_type)

        if header.version >= 17:
            writer.write_short(typ.script_type_index)

        if header.version >= 13:
            if (header.version < 16 and typ.class_id < 0) or (
                header.version >= 16 and typ.class_id == 114
            ):
                writer.write_bytes(typ.script_id)  # Hash128
            writer.write_bytes(typ.old_type_hash)  # Hash128

        if self._enable_type_tree:
            if header.version >= 12 or header.version == 10:
                self.save_type_tree5(typ.nodes, writer, typ.string_data)
            else:
                self.save_type_tree(typ.nodes, writer)

    def save_type_tree(self, nodes: list, writer: EndianBinaryWriter):
        for i, node in nodes:
            writer.write_string_to_null(node.type)
            writer.write_string_to_null(node.name)
            writer.write_int(node.byte_size)
            if self.header.version == 2:
                writer.write_int(node.variable_count)

            if self.header.version != 3:
                writer.write_int(node.index)

            writer.write_int(node.is_array)
            writer.write_int(node.version)
            if self.header.version != 3:
                writer.write_int(node.meta_flag)

            # calc children count
            children_count = 0
            for node2 in nodes[i + 1 :]:
                if node2.level == node.level:
                    break
                if node2.level == node.level - 1:
                    children_count += 1
            writer.write_int(children_count)

    def save_type_tree5(self, nodes: list, writer: EndianBinaryWriter, str_data=b""):
        # node count
        # stream buffer size
        # node data
        # string buffer

        string_buffer = EndianBinaryWriter()
        string_buffer.write(str_data)
        strings_values = [
            (node.type_str_offset, node.name_str_offset) for node in nodes
        ]

        # number of nodes
        writer.write_int(len(nodes))
        # string buffer size
        writer.write_int(string_buffer.Length)

        # nodes
        for i, node in enumerate(nodes):
            # version
            writer.write_u_short(node.version)
            # level
            writer.write_byte(node.level)
            # is array
            writer.write_boolean(node.is_array)
            # type str offfset
            writer.write_u_int(strings_values[i][0])
            # name str offset
            writer.write_u_int(strings_values[i][1])
            # byte size
            writer.write_int(node.byte_size)
            # index
            writer.write_int(node.index)
            # meta flag
            writer.write_int(node.meta_flag)
            # ref hash
            if self.header.version > 19:
                writer.write_u_long(node.ref_type_hash)

        # string buffer
        writer.write(string_buffer.bytes)

        if self.header.version >= 21:
            writer.write_bytes(b"\x00" * 4)


def read_string(string_buffer_reader: EndianBinaryReader, value: int) -> str:
    is_offset = (value & 0x80000000) == 0
    if is_offset:
        string_buffer_reader.Position = value
        return string_buffer_reader.read_string_to_null()

    offset = value & 0x7FFFFFFF
    if offset in CommonString:
        return CommonString[offset]

    return str(offset)




from .PPtr import PPtr
from .RuntimeAnimatorController import RuntimeAnimatorController


class AnimationClipOverride:
    def __init__(self, reader):
        self.m_OriginalClip = PPtr(reader)
        self.m_OverrideClip = PPtr(reader)


class AnimatorOverrideController(RuntimeAnimatorController):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_Controller = PPtr(reader)
        num_overrides = reader.read_int()
        self.m_Clips = [AnimationClipOverride(
            reader) for _ in range(num_overrides)]



from .NamedObject import NamedObject
from .PPtr import PPtr


class Font(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version
        if version >= (5, 5):  # 5.5 and up:
            self.m_LineSpacing = reader.read_float()
            self.m_DefaultMaterial = PPtr(reader)
            self.m_FontSize = reader.read_float()
            self.m_Texture = PPtr(reader)
            self.m_AsciiStartOffset = reader.read_int()
            self.m_Tracking = reader.read_float()
            self.m_CharacterSpacing = reader.read_int()
            self.m_CharacterPadding = reader.read_int()
            self.m_ConvertCase = reader.read_int()
            CharacterRects_size = reader.read_int()
            for i in range(CharacterRects_size):
                reader.Position += 44  # CharacterInfo data 41
            KerningValues_size = reader.read_int()
            for i in range(KerningValues_size):
                reader.Position += 8
            self.m_PixelScale = reader.read_float()
            FontData_size = reader.read_int()
            if FontData_size > 0:
                self.m_FontData = reader.read_bytes(FontData_size)
        else:
            self.m_AsciiStartOffset = reader.read_int()

            if version[:1] <= (3,):
                self.m_FontCountX = reader.read_int()
                self.m_FontCountY = reader.read_int()

            self.m_Kerning = reader.read_float()
            self.m_LineSpacing = reader.read_float()

            if version[:1] <= (3,):
                PerCharacterKerning_size = reader.read_int()
                for i in range(PerCharacterKerning_size):
                    first = reader.read_int()
                    second = reader.read_float()
            else:
                self.m_CharacterSpacing = reader.read_int()
                self.m_CharacterPadding = reader.read_int()

            self.m_ConvertCase = reader.read_int()
            self.m_DefaultMaterial = PPtr(reader)

            CharacterRects_size = reader.read_int()
            for i in range(CharacterRects_size):
                index = reader.read_int()
                # Rectf uv
                uvx = reader.read_float()
                uvy = reader.read_float()
                uvwidth = reader.read_float()
                uvheight = reader.read_float()
                # Rectf vert
                vertx = reader.read_float()
                verty = reader.read_float()
                vertwidth = reader.read_float()
                vertheight = reader.read_float()
                width = reader.read_float()

                if version >= (4,):
                    flipped = reader.read_boolean()
                    reader.align_stream()

            self.m_Texture = PPtr(reader)

            KerningValues_size = reader.read_int()
            for i in range(KerningValues_size):
                pairfirst = reader.read_short()
                pairsecond = reader.read_short()
                second = reader.read_float()

            if version[:1] <= (3,):
                self.m_GridFont = reader.read_boolean()
                reader.align_stream()
            else:
                self.m_PixelScale = reader.read_float()
            FontData_size = reader.read_int()
            if FontData_size > 0:
                self.m_FontData = reader.read_bytes(FontData_size)



from .Vector3 import Vector3


class Matrix4x4:
    M: list

    def __init__(self, values):
        if len(values) != 16:
            raise ValueError(
                "There must be sixteen and only sixteen input values for Matrix."
            )
        self.M = values

    def __getitem__(self, index):
        if isinstance(index, tuple):
            index = index[0] + index[1] * 4
        return self.M[index]

    def __setitem__(self, index, value):
        if isinstance(index, tuple):
            # row, column
            index = index[0] + index[1] * 4
        self.M[index] = value

    def __eq__(self, other):
        if not isinstance(other, Matrix4x4):
            return False
        print()

    def __mul__(lhs, rhs):
        res = Matrix4x4([0] * 16)
        res.M00 = (
            lhs.M00 * rhs.M00
            + lhs.M01 * rhs.M10
            + lhs.M02 * rhs.M20
            + lhs.M03 * rhs.M30
        )
        res.M01 = (
            lhs.M00 * rhs.M01
            + lhs.M01 * rhs.M11
            + lhs.M02 * rhs.M21
            + lhs.M03 * rhs.M31
        )
        res.M02 = (
            lhs.M00 * rhs.M02
            + lhs.M01 * rhs.M12
            + lhs.M02 * rhs.M22
            + lhs.M03 * rhs.M32
        )
        res.M03 = (
            lhs.M00 * rhs.M03
            + lhs.M01 * rhs.M13
            + lhs.M02 * rhs.M23
            + lhs.M03 * rhs.M33
        )

        res.M10 = (
            lhs.M10 * rhs.M00
            + lhs.M11 * rhs.M10
            + lhs.M12 * rhs.M20
            + lhs.M13 * rhs.M30
        )
        res.M11 = (
            lhs.M10 * rhs.M01
            + lhs.M11 * rhs.M11
            + lhs.M12 * rhs.M21
            + lhs.M13 * rhs.M31
        )
        res.M12 = (
            lhs.M10 * rhs.M02
            + lhs.M11 * rhs.M12
            + lhs.M12 * rhs.M22
            + lhs.M13 * rhs.M32
        )
        res.M13 = (
            lhs.M10 * rhs.M03
            + lhs.M11 * rhs.M13
            + lhs.M12 * rhs.M23
            + lhs.M13 * rhs.M33
        )

        res.M20 = (
            lhs.M20 * rhs.M00
            + lhs.M21 * rhs.M10
            + lhs.M22 * rhs.M20
            + lhs.M23 * rhs.M30
        )
        res.M21 = (
            lhs.M20 * rhs.M01
            + lhs.M21 * rhs.M11
            + lhs.M22 * rhs.M21
            + lhs.M23 * rhs.M31
        )
        res.M22 = (
            lhs.M20 * rhs.M02
            + lhs.M21 * rhs.M12
            + lhs.M22 * rhs.M22
            + lhs.M23 * rhs.M32
        )
        res.M23 = (
            lhs.M20 * rhs.M03
            + lhs.M21 * rhs.M13
            + lhs.M22 * rhs.M23
            + lhs.M23 * rhs.M33
        )

        res.M30 = (
            lhs.M30 * rhs.M00
            + lhs.M31 * rhs.M10
            + lhs.M32 * rhs.M20
            + lhs.M33 * rhs.M30
        )
        res.M31 = (
            lhs.M30 * rhs.M01
            + lhs.M31 * rhs.M11
            + lhs.M32 * rhs.M21
            + lhs.M33 * rhs.M31
        )
        res.M32 = (
            lhs.M30 * rhs.M02
            + lhs.M31 * rhs.M12
            + lhs.M32 * rhs.M22
            + lhs.M33 * rhs.M32
        )
        res.M33 = (
            lhs.M30 * rhs.M03
            + lhs.M31 * rhs.M13
            + lhs.M32 * rhs.M23
            + lhs.M33 * rhs.M33
        )

        return res

    @staticmethod
    def Scale(vector: Vector3):
        return Matrix4x4(
            [vector.X, 0, 0, 0, 0, vector.Y, 0, 0, 0, 0, vector.Z, 0, 0, 0, 0, 1]
        )

    @property
    def M00(self):
        return self.M[0]

    @M00.setter
    def M00(self, value):
        self.M[0] = value

    @property
    def M10(self):
        return self.M[1]

    @M10.setter
    def M10(self, value):
        self.M[1] = value

    @property
    def M20(self):
        return self.M[2]

    @M20.setter
    def M20(self, value):
        self.M[2] = value

    @property
    def M30(self):
        return self.M[3]

    @M30.setter
    def M30(self, value):
        self.M[3] = value

    @property
    def M01(self):
        return self.M[4]

    @M01.setter
    def M01(self, value):
        self.M[4] = value

    @property
    def M11(self):
        return self.M[5]

    @M11.setter
    def M11(self, value):
        self.M[5] = value

    @property
    def M21(self):
        return self.M[6]

    @M21.setter
    def M21(self, value):
        self.M[6] = value

    @property
    def M31(self):
        return self.M[7]

    @M31.setter
    def M31(self, value):
        self.M[7] = value

    @property
    def M02(self):
        return self.M[8]

    @M02.setter
    def M02(self, value):
        self.M[8] = value

    @property
    def M12(self):
        return self.M[9]

    @M12.setter
    def M12(self, value):
        self.M[9] = value

    @property
    def M22(self):
        return self.M[10]

    @M22.setter
    def M22(self, value):
        self.M[10] = value

    @property
    def M32(self):
        return self.M[11]

    @M32.setter
    def M32(self, value):
        self.M[11] = value

    @property
    def M03(self):
        return self.M[12]

    @M03.setter
    def M03(self, value):
        self.M[12] = value

    @property
    def M13(self):
        return self.M[13]

    @M13.setter
    def M13(self, value):
        self.M[13] = value

    @property
    def M23(self):
        return self.M[14]

    @M23.setter
    def M23(self, value):
        self.M[14] = value

    @property
    def M33(self):
        return self.M[15]

    @M33.setter
    def M33(self, value):
        self.M[15] = value



from .NamedObject import NamedObject
from ..streams import EndianBinaryWriter


class Texture(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        if self.version >= (2017, 3):  # 2017.3 and up
            self.m_ForcedFallbackFormat = reader.read_int()
            self.m_DownscaleFallback = reader.read_boolean()
            if self.version >= (2020,2): # 2020.2 and up
                self.m_IsAlphaChannelOptional = reader.read_boolean()
            reader.align_stream()

    def save(self, writer: EndianBinaryWriter):
        super().save(writer)
        if self.version >= (2017, 3):  # 2017.3 and up
            writer.write_int(self.m_ForcedFallbackFormat)
            writer.write_boolean(self.m_DownscaleFallback)
            if self.version >= (2020,2): # 2020.2 and up
                writer.write_boolean(self.m_IsAlphaChannelOptional)
            writer.align_stream()



import struct
import math

MaxValue = 65504.0
MinValue = -65504.0


def ToHalf(*args) -> float:
    """
    Converts the input into a half-float.
    Inputs:
        unsigned integer
        or
        buffer (bytes, buffer)
        offset
    """
    # int input -> pack as UInt16
    if len(args) == 1:
        data = struct.pack("H", args[0])
        val = struct.unpack("e", data)[0]
    # buffer input
    elif len(args) == 2:
        val = struct.unpack_from("e", args[0], args[1])[0]
    else:
        raise ValueError("Invalid amount of arguments")

    if math.isnan(val):
        # print('Nan')
        return 0
    elif math.isinf(val):
        return MaxValue

    return val


# #CONSTANTS
# Epsilon = ToHalf(0x0001)
# MaxValue = ToHalf(0x7bff)
# MinValue = ToHalf(0xfbff)
# NaN = ToHalf(0xfe00)
# NegativeInfinity = ToHalf(0xfc00)
# PositiveInfinity = ToHalf(0x7c00)

# class Float16Compressor:
#     def __init__(self):
#         self.temp = 0

#     def compress(self, float32):
#         F16_EXPONENT_BITS = 0x1F
#         F16_EXPONENT_SHIFT = 10
#         F16_EXPONENT_BIAS = 15
#         F16_MANTISSA_BITS = 0x3ff
#         F16_MANTISSA_SHIFT = (23 - F16_EXPONENT_SHIFT)
#         F16_MAX_EXPONENT = (F16_EXPONENT_BITS << F16_EXPONENT_SHIFT)

#         a = struct.pack('>f', float32)
#         b = binascii.hexlify(a)

#         f32 = int(b, 16)
#         f16 = 0
#         sign = (f32 >> 16) & 0x8000
#         exponent = ((f32 >> 23) & 0xff) - 127
#         mantissa = f32 & 0x007fffff

#         if exponent == 128:
#             f16 = sign | F16_MAX_EXPONENT
#             if mantissa:
#                 f16 |= (mantissa & F16_MANTISSA_BITS)
#         elif exponent > 15:
#             f16 = sign | F16_MAX_EXPONENT
#         elif exponent > -15:
#             exponent += F16_EXPONENT_BIAS
#             mantissa >>= F16_MANTISSA_SHIFT
#             f16 = sign | exponent << F16_EXPONENT_SHIFT | mantissa
#         else:
#             f16 = sign
#         return f16

#     def decompress(self, float16):
#         s = int((float16 >> 15) & 0x00000001)  # sign
#         e = int((float16 >> 10) & 0x0000001f)  # exponent
#         f = int(float16 & 0x000003ff)  # fraction

#         if e == 0:
#             if f == 0:
#                 return int(s << 31)
#             else:
#                 while not (f & 0x00000400):
#                     f = f << 1
#                     e -= 1
#                 e += 1
#                 f &= ~0x00000400
#         # print(s,e,f)
#         elif e == 31:
#             if f == 0:
#                 return int((s << 31) | 0x7f800000)
#             else:
#                 return int((s << 31) | 0x7f800000 | (f << 13))

#         e = e + (127 - 15)
#         f = f << 13
#         return int((s << 31) | (e << 23) | f)



from .Texture import Texture
from ..enums import TextureFormat
from ..export import Texture2DConverter
from ..helpers.ResourceReader import get_resource_data
from ..streams import EndianBinaryWriter
from PIL import Image
from io import BufferedIOBase, RawIOBase, IOBase


class Texture2D(Texture):
    @property
    def image(self):
        return Texture2DConverter.get_image_from_texture2d(self)

    @image.setter
    def image(self, img):
        # img is PIL.Image / image path / opened file
        # overwrite original image data with the RGB(A) image data and sets the correct new format
        if img is None:
            raise Exception("No image provided")

        if (
            isinstance(img, str)
            or isinstance(img, BufferedIOBase)
            or isinstance(img, RawIOBase)
            or isinstance(img, IOBase)
        ):
            img = Image.open(img)

        img_data, tex_format = Texture2DConverter.image_to_texture2d(
            img, self.m_TextureFormat
        )

        # disable mipmaps as we don't store them ourselves by default
        if self.version[:2] < (5, 2):  # 5.2 down
            self.m_MipMap = False
        else:
            self.m_MipCount = 1

        self.image_data = img_data
        self.m_MipCount = 1
        # width * height * channel count
        self.m_CompleteImageSize = len(
            img_data
        )  # img.width * img.height * len(img.getbands())
        self.m_TextureFormat = tex_format

    @property
    def image_data(self):
        return self._image_data

    def reset_streamdata(self):
        if not self.m_StreamData:
            return
        self.m_StreamData.offset = 0
        self.m_StreamData.size = 0
        self.m_StreamData.path = ""

    @image_data.setter
    def image_data(self, data: bytes):
        self._image_data = data
        # ignore writing to cab for now until it's more stable
        # if self.version >= (5, 3) and self.m_StreamData.path:
        #     cab = self.assets_file.get_writeable_cab()
        #     if cab:
        #         self.m_StreamData.offset = cab.Position
        #         cab.write(data)
        #         self.m_StreamData.size = len(data)
        #         self.m_StreamData.path = cab.path
        #     else:
        #         self.reset_streamdata()
        self.reset_streamdata()

    def set_image(
        self,
        img,
        target_format: TextureFormat = None,
        in_cab: bool = False,
        mipmap_count: int = 1,
    ):
        if img is None:
            raise Exception("No image provided")
        if not target_format:
            target_format = self.m_TextureFormat

        img_data, tex_format = Texture2DConverter.image_to_texture2d(img, target_format)
        if mipmap_count > 1:
            width = self.m_Width
            height = self.m_Height
            re_img = img
            for i in range(mipmap_count - 1):
                width //= 2
                height //= 2
                if width < 4 or height < 4:
                    mipmap_count = i + 1
                    break
                re_img = re_img.resize((width, height), Image.BICUBIC)
                img_data += Texture2DConverter.image_to_texture2d(
                    re_img, target_format
                )[0]

        if self.version[:2] < (5, 2):  # 5.2 down
            self.m_MipMap = mipmap_count > 1
        else:
            self.m_MipCount = mipmap_count

        if in_cab:
            self.image_data = img_data
        else:
            self._image_data = img_data
            self.reset_streamdata()

        # width * height * channel count
        self.m_CompleteImageSize = len(
            img_data
        )  # img.width * img.height * len(img.getbands())
        self.m_TextureFormat = tex_format

    def __init__(self, reader):
        super().__init__(reader=reader)
        version = self.version

        self.m_Width = reader.read_int()
        self.m_Height = reader.read_int()
        self.m_CompleteImageSize = reader.read_int()
        if version >= (2020,):  # 2020.1 and up
            self.m_MipsStripped = reader.read_int()
        self.m_TextureFormat = TextureFormat(reader.read_int())
        if version[:2] < (5, 2):  # 5.2 down
            self.m_MipMap = reader.read_boolean()
        else:
            self.m_MipCount = reader.read_int()

        if version >= (2, 6):  # 2.6 and up
            self.m_IsReadable = reader.read_boolean()  # 2.6 and up
        if version >= (2020,):  # 2020.1 and up
            self.m_IsPreProcessed = reader.read_boolean()
        if version >= (2019, 3):  # 2019.3 and up
            self.m_IgnoreMasterTextureLimit = reader.read_boolean()
        if (3,) <= version[:2] <= (5, 4):  # 3.0 - 5.4
            self.m_ReadAllowed = reader.read_boolean()
        if version >= (2018, 2):  # 2018.2 and up
            self.m_StreamingMipmaps = reader.read_boolean()

        reader.align_stream()
        if version >= (2018, 2):  # 2018.2 and up
            self.m_StreamingMipmapsPriority = reader.read_int()
        self.m_ImageCount = reader.read_int()
        self.m_TextureDimension = reader.read_int()
        self.m_TextureSettings = GLTextureSettings(reader, version)
        if version >= (3,):  # 3.0 and up
            self.m_LightmapFormat = reader.read_int()
        if version >= (3, 5):  # 3.5 and up
            self.m_ColorSpace = reader.read_int()
        if version >= (2020, 2):  # 2020.2 and up
            self.m_PlatformBlob = reader.read_byte_array()
            reader.align_stream()

        image_data_size = reader.read_int()
        self._image_data = b""

        if image_data_size != 0:
            self._image_data = reader.read_bytes(image_data_size)

        self.m_StreamData = None
        if version >= (5, 3):  # 5.3 and up
            # always read the StreamingInfo for resaving
            self.m_StreamData = StreamingInfo(reader, version)
            if image_data_size == 0 and self.m_StreamData.path:
                self._image_data = get_resource_data(
                    self.m_StreamData.path,
                    self.assets_file,
                    self.m_StreamData.offset,
                    self.m_StreamData.size,
                )

    def save(self, writer: EndianBinaryWriter = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        version = self.version

        super().save(writer)
        writer.write_int(self.m_Width)
        writer.write_int(self.m_Height)
        writer.write_int(self.m_CompleteImageSize)
        if version >= (2020,):  # 2020.1 and up
            writer.write_int(self.m_MipsStripped)
        writer.write_int(self.m_TextureFormat.value)
        if version[:2] < (5, 2):  # 5.2 down
            writer.write_boolean(self.m_MipMap)
        else:
            writer.write_int(self.m_MipCount)

        if version >= (2, 6):  # 2.6 and up
            writer.write_boolean(self.m_IsReadable)  # 2.6 and up
        if version >= (2020,):  # 2020.1 and up
            writer.write_boolean(self.m_IsPreProcessed)
        if version >= (2019, 3):  # 2019.3 and up
            writer.write_boolean(self.m_IgnoreMasterTextureLimit)
        if (3,) <= version[:2] <= (5, 4):  # 3.0 - 5.4
            writer.write_boolean(self.m_ReadAllowed)  # 3.0 - 5.4
        if version >= (2018, 2):  # 2018.2 and up
            writer.write_boolean(self.m_StreamingMipmaps)

        writer.align_stream()
        if version >= (2018, 2):  # 2018.2 and up
            writer.write_int(self.m_StreamingMipmapsPriority)
        writer.write_int(self.m_ImageCount)
        writer.write_int(self.m_TextureDimension)
        self.m_TextureSettings.save(writer, version)
        if version >= (3,):  # 3.0 and up
            writer.write_int(self.m_LightmapFormat)
        if version >= (3, 5):  # 3.5 and up
            writer.write_int(self.m_ColorSpace)
        if version >= (2020, 2):  # 2020.2 and up
            writer.write_byte_array(self.m_PlatformBlob)
            writer.align_stream()

        if version[:2] < (5, 3):
            # version without m_StreamData
            writer.write_int(len(self.image_data))
            writer.write_bytes(self.image_data)
        else:
            # decide if m_StreamData is used
            if self.m_StreamData.path:
                # used -> don't save the image_data
                writer.write_int(0)
            else:
                writer.write_int(len(self.image_data))
                writer.write_bytes(self.image_data)

            self.m_StreamData.save(writer, version)

        self.set_raw_data(writer.bytes)


class StreamingInfo:
    offset: int
    size: int
    path: str

    def __init__(self, reader, version):
        if version >= (2020,):  # 2020.1 and up
            self.offset = reader.read_u_long()
        else:
            self.offset = reader.read_u_int()
        self.size = reader.read_u_int()
        self.path = reader.read_aligned_string()

    def save(self, writer, version):
        if version >= (2020,):  # 2020.1 and up
            writer.write_u_long(self.offset)
        else:
            writer.write_u_int(self.offset)
        writer.write_int(self.size)
        writer.write_aligned_string(self.path)


class GLTextureSettings:
    def __init__(self, reader, version):
        self.m_FilterMode = reader.read_int()
        self.m_Aniso = reader.read_int()
        self.m_MipBias = reader.read_float()
        self.m_WrapMode = reader.read_int()  # m_WrapU
        if version >= (2017,):  # 2017.x and up
            self.m_WrapV = reader.read_int()
            self.m_WrapW = reader.read_int()

    def save(self, writer, version):
        writer.write_int(self.m_FilterMode)
        writer.write_int(self.m_Aniso)
        writer.write_float(self.m_MipBias)
        writer.write_int(self.m_WrapMode)  # m_WrapU
        if version >= (2017,):  # 2017.x and up
            writer.write_int(self.m_WrapV)
            writer.write_int(self.m_WrapW)



# https://docs.unity3d.com/Manual/ClassIDReference.html
from .ExtendableEnum import ExtendableEnum


class ClassIDType(ExtendableEnum):
    UnknownType = -1
    Object = 0
    GameObject = 1
    Component = 2
    LevelGameManager = 3
    Transform = 4
    TimeManager = 5
    GlobalGameManager = 6
    Behaviour = 8
    GameManager = 9
    AudioManager = 11
    ParticleAnimator = 12
    InputManager = 13
    EllipsoidParticleEmitter = 15
    Pipeline = 17
    EditorExtension = 18
    Physics2DSettings = 19
    Camera = 20
    Material = 21
    MeshRenderer = 23
    Renderer = 25
    ParticleRenderer = 26
    Texture = 27
    Texture2D = 28
    OcclusionCullingSettings = 29
    GraphicsSettings = 30
    MeshFilter = 33
    OcclusionPortal = 41
    Mesh = 43
    Skybox = 45
    QualitySettings = 47
    Shader = 48
    TextAsset = 49
    Rigidbody2D = 50
    Physics2DManager = 51
    Collider2D = 53
    Rigidbody = 54
    PhysicsManager = 55
    Collider = 56
    Joint = 57
    CircleCollider2D = 58
    HingeJoint = 59
    PolygonCollider2D = 60
    BoxCollider2D = 61
    PhysicsMaterial2D = 62
    MeshCollider = 64
    BoxCollider = 65
    CompositeCollider2D = 66
    EdgeCollider2D = 68
    CapsuleCollider2D = 70
    ComputeShader = 72
    AnimationClip = 74
    ConstantForce = 75
    WorldParticleCollider = 76
    TagManager = 78
    AudioListener = 81
    AudioSource = 82
    AudioClip = 83
    RenderTexture = 84
    CustomRenderTexture = 86
    MeshParticleEmitter = 87
    ParticleEmitter = 88
    Cubemap = 89
    Avatar = 90
    AnimatorController = 91
    GUILayer = 92
    RuntimeAnimatorController = 93
    ScriptMapper = 94
    Animator = 95
    TrailRenderer = 96
    DelayedCallManager = 98
    TextMesh = 102
    RenderSettings = 104
    Light = 108
    CGProgram = 109
    BaseAnimationTrack = 110
    Animation = 111
    MonoBehaviour = 114
    MonoScript = 115
    MonoManager = 116
    Texture3D = 117
    NewAnimationTrack = 118
    Projector = 119
    LineRenderer = 120
    Flare = 121
    Halo = 122
    LensFlare = 123
    FlareLayer = 124
    HaloLayer = 125
    NavMeshProjectSettings = 126
    HaloManager = 127
    Font = 128
    PlayerSettings = 129
    NamedObject = 130
    GUITexture = 131
    GUIText = 132
    GUIElement = 133
    PhysicMaterial = 134
    SphereCollider = 135
    CapsuleCollider = 136
    SkinnedMeshRenderer = 137
    FixedJoint = 138
    RaycastCollider = 140
    BuildSettings = 141
    AssetBundle = 142
    CharacterController = 143
    CharacterJoint = 144
    SpringJoint = 145
    WheelCollider = 146
    ResourceManager = 147
    NetworkView = 148
    NetworkManager = 149
    PreloadData = 150
    MovieTexture = 152
    ConfigurableJoint = 153
    TerrainCollider = 154
    MasterServerInterface = 155
    TerrainData = 156
    LightmapSettings = 157
    WebCamTexture = 158
    EditorSettings = 159
    InteractiveCloth = 160
    ClothRenderer = 161
    EditorUserSettings = 162
    SkinnedCloth = 163
    AudioReverbFilter = 164
    AudioHighPassFilter = 165
    AudioChorusFilter = 166
    AudioReverbZone = 167
    AudioEchoFilter = 168
    AudioLowPassFilter = 169
    AudioDistortionFilter = 170
    SparseTexture = 171
    AudioBehaviour = 180
    AudioFilter = 181
    WindZone = 182
    Cloth = 183
    SubstanceArchive = 184
    ProceduralMaterial = 185
    ProceduralTexture = 186
    Texture2DArray = 187
    CubemapArray = 188
    OffMeshLink = 191
    OcclusionArea = 192
    Tree = 193
    NavMeshObsolete = 194
    NavMeshAgent = 195
    NavMeshSettings = 196
    LightProbesLegacy = 197
    ParticleSystem = 198
    ParticleSystemRenderer = 199
    ShaderVariantCollection = 200
    LODGroup = 205
    BlendTree = 206
    Motion = 207
    NavMeshObstacle = 208
    SortingGroup = 210
    SpriteRenderer = 212
    Sprite = 213
    CachedSpriteAtlas = 214
    ReflectionProbe = 215
    ReflectionProbes = 216
    Terrain = 218
    LightProbeGroup = 220
    AnimatorOverrideController = 221
    CanvasRenderer = 222
    Canvas = 223
    RectTransform = 224
    CanvasGroup = 225
    BillboardAsset = 226
    BillboardRenderer = 227
    SpeedTreeWindAsset = 228
    AnchoredJoint2D = 229
    Joint2D = 230
    SpringJoint2D = 231
    DistanceJoint2D = 232
    HingeJoint2D = 233
    SliderJoint2D = 234
    WheelJoint2D = 235
    ClusterInputManager = 236
    BaseVideoTexture = 237
    NavMeshData = 238
    AudioMixer = 240
    AudioMixerController = 241
    AudioMixerGroupController = 243
    AudioMixerEffectController = 244
    AudioMixerSnapshotController = 245
    PhysicsUpdateBehaviour2D = 246
    ConstantForce2D = 247
    Effector2D = 248
    AreaEffector2D = 249
    PointEffector2D = 250
    PlatformEffector2D = 251
    SurfaceEffector2D = 252
    BuoyancyEffector2D = 253
    RelativeJoint2D = 254
    FixedJoint2D = 255
    FrictionJoint2D = 256
    TargetJoint2D = 257
    LightProbes = 258
    LightProbeProxyVolume = 259
    SampleClip = 271
    AudioMixerSnapshot = 272
    AudioMixerGroup = 273
    NScreenBridge = 280
    AssetBundleManifest = 290
    UnityAdsManager = 292
    RuntimeInitializeOnLoadManager = 300
    CloudWebServicesManager = 301
    UnityAnalyticsManager = 303
    CrashReportManager = 304
    PerformanceReportingManager = 305
    UnityConnectSettings = 310
    AvatarMask = 319
    PlayableDirector = 320
    VideoPlayer = 328
    VideoClip = 329
    ParticleSystemForceField = 330
    SpriteMask = 331
    WorldAnchor = 362
    OcclusionCullingData = 363
    SmallestEditorClassID = 1000
    PrefabInstance = 1001
    EditorExtensionImpl = 1002
    AssetImporter = 1003
    AssetDatabaseV1 = 1004
    Mesh3DSImporter = 1005
    TextureImporter = 1006
    ShaderImporter = 1007
    ComputeShaderImporter = 1008
    AudioImporter = 1020
    HierarchyState = 1026
    GUIDSerializer = 1027
    AssetMetaData = 1028
    DefaultAsset = 1029
    DefaultImporter = 1030
    TextScriptImporter = 1031
    SceneAsset = 1032
    NativeFormatImporter = 1034
    MonoImporter = 1035
    AssetServerCache = 1037
    LibraryAssetImporter = 1038
    ModelImporter = 1040
    FBXImporter = 1041
    TrueTypeFontImporter = 1042
    MovieImporter = 1044
    EditorBuildSettings = 1045
    DDSImporter = 1046
    InspectorExpandedState = 1048
    AnnotationManager = 1049
    PluginImporter = 1050
    EditorUserBuildSettings = 1051
    PVRImporter = 1052
    ASTCImporter = 1053
    KTXImporter = 1054
    IHVImageFormatImporter = 1055
    AnimatorStateTransition = 1101
    AnimatorState = 1102
    HumanTemplate = 1105
    AnimatorStateMachine = 1107
    PreviewAnimationClip = 1108
    AnimatorTransition = 1109
    SpeedTreeImporter = 1110
    AnimatorTransitionBase = 1111
    SubstanceImporter = 1112
    LightmapParameters = 1113
    LightingDataAsset = 1120
    GISRaster = 1121
    GISRasterImporter = 1122
    CadImporter = 1123
    SketchUpImporter = 1124
    BuildReport = 1125
    PackedAssets = 1126
    VideoClipImporter = 1127
    ActivationLogComponent = 2000
    int = 100000
    bool = 100001
    float = 100002
    MonoObject = 100003
    Collision = 100004
    Vector3f = 100005
    RootMotionData = 100006
    Collision2D = 100007
    AudioMixerLiveUpdateFloat = 100008
    AudioMixerLiveUpdateBool = 100009
    Polygon2D = 100010
    void = 100011
    TilemapCollider2D = 19719996
    AssetImporterLog = 41386430
    VFXRenderer = 73398921
    SerializableManagedRefTestClass = 76251197
    Grid = 156049354
    ScenesUsingAssets = 156483287
    ArticulationBody = 171741748
    Preset = 181963792
    EmptyObject = 277625683
    IConstraint = 285090594
    TestObjectWithSpecialLayoutOne = 293259124
    AssemblyDefinitionReferenceImporter = 294290339
    SiblingDerived = 334799969
    TestObjectWithSerializedMapStringNonAlignedStruct = 342846651
    SubDerived = 367388927
    AssetImportInProgressProxy = 369655926
    PluginBuildInfo = 382020655
    EditorProjectAccess = 426301858
    PrefabImporter = 468431735
    TestObjectWithSerializedArray = 478637458
    TestObjectWithSerializedAnimationCurve = 478637459
    TilemapRenderer = 483693784
    ScriptableCamera = 488575907
    SpriteAtlasAsset = 612988286
    SpriteAtlasDatabase = 638013454
    AudioBuildInfo = 641289076
    CachedSpriteAtlasRuntimeData = 644342135
    RendererFake = 646504946
    AssemblyDefinitionReferenceAsset = 662584278
    BuiltAssetBundleInfoSet = 668709126
    SpriteAtlas = 687078895
    RayTracingShaderImporter = 747330370
    RayTracingShader = 825902497
    LightingSettings = 850595691
    PlatformModuleSetup = 877146078
    VersionControlSettings = 890905787
    AimConstraint = 895512359
    VFXManager = 937362698
    VisualEffectSubgraph = 994735392
    VisualEffectSubgraphOperator = 994735403
    VisualEffectSubgraphBlock = 994735404
    LocalizationImporter = 1027052791
    Derived = 1091556383
    PropertyModificationsTargetTestObject = 1111377672
    ReferencesArtifactGenerator = 1114811875
    AssemblyDefinitionAsset = 1152215463
    SceneVisibilityState = 1154873562
    LookAtConstraint = 1183024399
    SpriteAtlasImporter = 1210832254
    MultiArtifactTestImporter = 1223240404
    GameObjectRecorder = 1268269756
    LightingDataAssetParent = 1325145578
    PresetManager = 1386491679
    TestObjectWithSpecialLayoutTwo = 1392443030
    StreamingManager = 1403656975
    LowerResBlitTexture = 1480428607
    StreamingController = 1542919678
    RenderPassAttachment = 1571458007
    TestObjectVectorPairStringBool = 1628831178
    GridLayout = 1742807556
    AssemblyDefinitionImporter = 1766753193
    ParentConstraint = 1773428102
    FakeComponent = 1803986026
    PositionConstraint = 1818360608
    RotationConstraint = 1818360609
    ScaleConstraint = 1818360610
    Tilemap = 1839735485
    PackageManifest = 1896753125
    PackageManifestImporter = 1896753126
    TerrainLayer = 1953259897
    SpriteShapeRenderer = 1971053207
    NativeObjectType = 1977754360
    TestObjectWithSerializedMapStringBool = 1981279845
    SerializableManagedHost = 1995898324
    VisualEffectAsset = 2058629509
    VisualEffectImporter = 2058629510
    VisualEffectResource = 2058629511
    VisualEffectObject = 2059678085
    VisualEffect = 2083052967
    LocalizationAsset = 2083778819
    ScriptedImporter = 2089858483



from ctypes import *
from enum import Enum
import os
import platform
from UnityPy.streams import EndianBinaryWriter

_dll = None


def load_fmod_library():
    global _dll
    if _dll is not None:
        return

    ROOT = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

    # determine system - Windows, Darwin, Linux, Android
    system = platform.system()
    if system == "Linux" and "ANDROID_BOOTLOGO" in os.environ:
        system = "Android"
    # determine architecture
    machine = platform.machine()
    arch = platform.architecture()[0]

    if system in ["Windows", "Darwin"]:
        if arch == "32bit":
            arch = "x86"
        elif arch == "64bit":
            arch = "x64"
    elif system == "Linux":
        # Raspberry Pi and Linux on arm projects
        if "arm" in machine:
            if arch == "32bit":
                arch = "armhf" if machine.endswith("l") else "arm"
            elif arch == "64bit":
                # Raise an exception for now; Once it gets supported by FMOD we can just modify the code here
                _dll = False
                raise NotImplementedError("ARM64 not supported by FMOD.\nUse a 32bit python version.")
        elif arch == "32bit":
            arch = "x86"
        elif arch == "64bit":
            arch = "x86_64"
    else:
        _dll = False
        raise NotImplementedError("Couldn't find a correct FMOD library for your system ({system} - {arch}).")
    try:
        # build path and load library
        LIB_PATH = os.path.join(ROOT, "lib", "FMOD", system, arch)
        if system == 'Windows':
            _dll = WinDLL(os.path.join(LIB_PATH, "fmod.dll"))
        elif system in ["Linux", "Android"]:
            _dll = CDLL(os.path.join(LIB_PATH, "libfmod.so"))
        elif system == "Darwin":
            _dll = CDLL(os.path.join(LIB_PATH, "libfmod.dylib"))
    except Exception as e:
        raise ImportError(
            f"Failed to import the fmod library - Exception: {e}.\
            If you want to export AudioClips, you have to set UnityPy.export.AudioClipConverter._dll yourself."
        )


def extract_audioclip_samples(audio) -> dict:
    """extracts all the samples from an AudioClip
        :param audio: AudioClip
        :type audio: AudioClip
        :return: {filename : sample(bytes)}
        :rtype: dict
    """
    if not audio.m_AudioData:
        # eg. StreamedResource not available
        return {}

    magic = memoryview(audio.m_AudioData)[:4]
    if magic == b'OggS':
        return {'%s.ogg' % audio.name: audio.m_AudioData}
    elif magic == b'RIFF':
        return {'%s.wav' % audio.name: audio.m_AudioData}
    return dump_samples(audio)


def dump_samples(clip):
    if _dll is None:
        load_fmod_library()
    if not _dll:
        return {}
    # init system
    # system = pyfmodex.System()
    sys_ptr = c_void_p()
    ckresult(_dll.FMOD_System_Create(byref(sys_ptr)))
    # system.init(1, INIT_FLAGS.NORMAL, None)
    ckresult(_dll.FMOD_System_Init(sys_ptr, clip.m_Channels, None, None))

    # get sound
    exinfo = byref(CREATESOUNDEXINFO(length=clip.m_Size))
    # sound = system.create_sound(bytes(clip.m_AudioData),mode=MODE.OPENMEMORY,exinfo=exinfo)
    snd_ptr = c_void_p()
    ckresult(_dll.FMOD_System_CreateSound(sys_ptr, bytes(
        clip.m_AudioData), 0x00000800, exinfo, byref(snd_ptr)))
    sound = Sound(snd_ptr)
    # iterate over subsounds
    samples = {}
    for i in range(sound.num_subsounds):
        if i > 0:
            filename = "%s-%i.wav" % (clip.name, i)
        else:
            filename = "%s.wav" % clip.name
        subsound = sound.get_subsound(i)
        samples[filename] = subsound_to_wav(subsound)
        subsound.release()

    sound.release()
    # system.release()
    ckresult(_dll.FMOD_System_Release(sys_ptr))
    return samples


def subsound_to_wav(subsound):
    # get sound settings
    length = subsound.get_length(0x00000004)  # TIMEUNIT.PCMBYTES
    channels = subsound.format.channels
    bits = subsound.format.bits
    sample_rate = int(subsound.default_frequency)

    # write to buffer
    w = EndianBinaryWriter(endian="<")
    # riff chucnk
    w.write(b"RIFF")
    w.write_int(length + 36)  # sizeof(FmtChunk) + sizeof(RiffChunk) + length
    w.write(b"WAVE")
    # fmt chunck
    w.write(b"fmt ")
    w.write_int(16)  # sizeof(FmtChunk) - sizeof(RiffChunk)
    w.write_short(1)
    w.write_short(channels)
    w.write_int(sample_rate)
    w.write_int(sample_rate * channels * bits // 8)
    w.write_short(channels * bits // 8)
    w.write_short(bits)
    # data chunck
    w.write(b"data")
    w.write_int(length)
    # data
    lock = subsound.lock(0, length)
    for ptr, length in lock:
        ptr_data = string_at(ptr, length.value)
        w.write(ptr_data)
    subsound.unlock(*lock)
    return w.bytes


# following code is copied from
# https://github.com/tyrylu/pyfmodex
# pyfmodex can't be used by itself
# because it would require adding the libs to the path first

class CREATESOUNDEXINFO(Structure):
    _fields_ = [("cbsize", c_int), ("length", c_uint)]

    def __init__(self, *args, **kwargs):
        Structure.__init__(self, *args, **kwargs)
        self.cbsize = sizeof(self)


class so:
    def __init__(self, **kwargs):
        self.__dict__.update(**kwargs)


class Sound(object):
    def __init__(self, ptr):
        """Constructor.
        :param ptr: The pointer representing this object.
        """
        self._ptr = ptr

    def _call_fmod(self, funcname, *args):
        result = getattr(_dll, funcname)(self._ptr, *args)
        ckresult(result)

    @property
    def num_subsounds(self):
        num = c_int()
        self._call_fmod("FMOD_Sound_GetNumSubSounds", byref(num))
        return num.value

    def get_subsound(self, index):
        sh_ptr = c_void_p()
        self._call_fmod("FMOD_Sound_GetSubSound", index, byref(sh_ptr))
        return Sound(sh_ptr)

    def get_length(self, ltype):
        len = c_uint()
        self._call_fmod("FMOD_Sound_GetLength", byref(len), int(ltype))
        return len.value

    @property
    def format(self):
        type = c_int()
        format = c_int()
        channels = c_int()
        bits = c_int()
        self._call_fmod("FMOD_Sound_GetFormat", byref(type),
                        byref(format), byref(channels), byref(bits))
        return so(type=type.value, format=format.value, channels=channels.value, bits=bits.value)

    @property
    def default_frequency(self):
        freq = c_float()
        pri = c_int()
        self._call_fmod("FMOD_Sound_GetDefaults", byref(freq), byref(pri))
        return freq.value

    def lock(self, offset, length):
        ptr1 = c_void_p()
        len1 = c_uint()
        ptr2 = c_void_p()
        len2 = c_uint()
        ckresult(_dll.FMOD_Sound_Lock(self._ptr, offset, length,
                                      byref(ptr1), byref(ptr2), byref(len1), byref(len2)))
        return (ptr1, len1), (ptr2, len2)

    def release(self):
        self._call_fmod("FMOD_Sound_Release")

    def unlock(self, i1, i2):
        """I1 and I2 are tuples of form (ptr, len)."""
        ckresult(_dll.FMOD_Sound_Unlock(self._ptr, i1[0], i2[0], i1[1], i2[1]))


def ckresult(result):
    result = RESULT(result)
    if result is not RESULT.OK:
        raise FmodError(result)


class FmodError(Exception):
    def __init__(self, result):
        self.result = result
        self.message = result.name.replace("_", " ")

    def __str__(self):
        return self.message


class RESULT(Enum):
    OK = 0
    BADCOMMAND = 1
    CHANNEL_ALLOC = 2
    CHANNEL_STOLEN = 3
    DMA = 4
    DSP_CONNECTION = 5
    DSP_DONTPROCESS = 6
    DSP_FORMAT = 7
    DSP_INUSE = 8
    DSP_NOTFOUND = 9
    DSP_RESERVED = 10
    DSP_SILENCE = 11
    DSP_TYPE = 12
    FILE_BAD = 13
    FILE_COULDNOTSEEK = 14
    FILE_DISKEJECTED = 15
    FILE_EOF = 16
    FILE_ENDOFDATA = 17
    FILE_NOTFOUND = 18
    FORMAT = 19
    HEADER_MISMATCH = 20
    HTTP = 21
    HTTP_ACCESS = 22
    HTTP_PROXY_AUTH = 23
    HTTP_SERVER_ERROR = 24
    HTTP_TIMEOUT = 25
    INITIALIZATION = 26
    INITIALIZED = 27
    INTERNAL = 28
    INVALID_FLOAT = 29
    INVALID_HANDLE = 30
    INVALID_PARAM = 31
    INVALID_POSITION = 32
    INVALID_SPEAKER = 33
    INVALID_SYNCPOINT = 34
    INVALID_THREAD = 35
    INVALID_VECTOR = 36
    MAXAUDIBLE = 37
    MEMORY = 38
    MEMORY_CANTPOINT = 39
    NEEDS3D = 40
    NEEDSHARDWARE = 41
    NET_CONNECT = 42
    NET_SOCKET_ERROR = 43
    NET_URL = 44
    NET_WOULD_BLOCK = 45
    NOTREADY = 46
    OUTPUT_ALLOCATED = 47
    OUTPUT_CREATEBUFFER = 48
    OUTPUT_DRIVERCALL = 49
    OUTPUT_FORMAT = 50
    OUTPUT_INIT = 51
    OUTPUT_NODRIVERS = 52
    PLUGIN = 53
    PLUGIN_MISSING = 54
    PLUGIN_RESOURCE = 55
    PLUGIN_VERSION = 56
    RECORD = 57
    REVERB_CHANNELGROUP = 58
    REVERB_INSTANCE = 59
    SUBSOUNDS = 60
    SUBSOUND_ALLOCATED = 61
    SUBSOUND_CANTMOVE = 62
    TAGNOTFOUND = 63
    TOOMANYCHANNELS = 64
    TRUNCATED = 65
    UNIMPLEMENTED = 66
    UNINITIALIZED = 67
    UNSUPPORTED = 68
    VERSION = 69
    EVENT_ALREADY_LOADED = 70
    EVENT_LIVEUPDATE_BUSY = 71
    EVENT_LIVEUPDATE_MISMATCH = 72
    EVENT_LIVEUPDATE_TIMEOUT = 73
    EVENT_NOTFOUND = 74
    STUDIO_UNINITIALIZED = 75
    STUDIO_NOT_LOADED = 76
    INVALID_STRING = 77
    ALREADY_LOCKED = 78
    NOT_LOCKED = 79
    RECORD_DISCONNECTED = 80
    TOOMANYSAMPLES = 81



from .NamedObject import NamedObject
from ..streams import EndianBinaryWriter


class TextAsset(NamedObject):
    def __init__(self, reader):
        super().__init__(reader=reader)
        self.m_Script = reader.read_bytes(reader.read_int())

    @property
    def script(self):
        # required for backward compatibility
        return self.m_Script

    @script.setter
    def script(self, value):
        self.m_Script = value

    @property
    def text(self):
        return bytes(self.script).decode("utf8")

    @text.setter
    def text(self, val):
        self.script = val.encode("utf8")

    def save(self, writer: EndianBinaryWriter = None):
        if writer is None:
            writer = EndianBinaryWriter(endian=self.reader.endian)
        super().save(writer)
        writer.write_int(len(self.m_Script))
        writer.write_bytes(self.m_Script)
        writer.align_stream()

        self.set_raw_data(writer.bytes)



from .ExtendableEnum import ExtendableEnum


class BuildTarget(ExtendableEnum):
    UnknownPlatform = 3716
    DashboardWidget = 1
    StandaloneOSX = 2
    StandaloneOSXPPC = 3
    StandaloneOSXIntel = 4
    StandaloneWindows = 5
    WebPlayer = 6
    WebPlayerStreamed = 7
    Wii = 8
    iOS = 9
    PS3 = 10
    XBOX360 = 11
    Android = 13
    StandaloneGLESEmu = 14
    NaCl = 16
    StandaloneLinux = 17
    FlashPlayer = 18
    StandaloneWindows64 = 19
    WebGL = 20
    WSAPlayer = 21
    StandaloneLinux64 = 24
    StandaloneLinuxUniversal = 25
    WP8Player = 26
    StandaloneOSXIntel64 = 27
    BlackBerry = 28
    Tizen = 29
    PSP2 = 30
    PS4 = 31
    PSM = 32
    XboxOne = 33
    SamsungTV = 34
    N3DS = 35
    WiiU = 36
    tvOS = 37
    Switch = 38
    NoTarget = -2



from enum import IntEnum


class ShaderGpuProgramType(IntEnum):
    kShaderGpuProgramUnknown = 0
    kShaderGpuProgramGLLegacy = 1
    kShaderGpuProgramGLES31AEP = 2
    kShaderGpuProgramGLES31 = 3
    kShaderGpuProgramGLES3 = 4
    kShaderGpuProgramGLES = 5
    kShaderGpuProgramGLCore32 = 6
    kShaderGpuProgramGLCore41 = 7
    kShaderGpuProgramGLCore43 = 8
    kShaderGpuProgramDX9VertexSM20 = 9
    kShaderGpuProgramDX9VertexSM30 = 10
    kShaderGpuProgramDX9PixelSM20 = 11
    kShaderGpuProgramDX9PixelSM30 = 12
    kShaderGpuProgramDX10Level9Vertex = 13
    kShaderGpuProgramDX10Level9Pixel = 14
    kShaderGpuProgramDX11VertexSM40 = 15
    kShaderGpuProgramDX11VertexSM50 = 16
    kShaderGpuProgramDX11PixelSM40 = 17
    kShaderGpuProgramDX11PixelSM50 = 18
    kShaderGpuProgramDX11GeometrySM40 = 19
    kShaderGpuProgramDX11GeometrySM50 = 20
    kShaderGpuProgramDX11HullSM50 = 21
    kShaderGpuProgramDX11DomainSM50 = 22
    kShaderGpuProgramMetalVS = 23
    kShaderGpuProgramMetalFS = 24
    kShaderGpuProgramSPIRV = 25
    kShaderGpuProgramConsoleVS = 26
    kShaderGpuProgramConsoleFS = 27
    kShaderGpuProgramConsoleHS = 28
    kShaderGpuProgramConsoleDS = 29
    kShaderGpuProgramConsoleGS = 30
    kShaderGpuProgramRayTracing = 31
